%% DESCRIPTION
% 12/01/2022 Eigen study of fluid rectangle
% R. Liupekevicius Carnielli
% the goal of this script is to compute a simple eigenstudy and validate it
% against comsol simulation. 
%
%
% based com "compression_solution.m" from ACCM
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  - open: pressure = 0
%  - hard: velocity = 0 ( or pressure gradient =0)
%
%          ___(hard)__
%        /|           |\
%  (open)/|           |\ (open)
%        /|           |\
%        /|___(hard)__|\
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
% What are the low frequency modes of this cavity?
%
% Make sure you also use same element type in comsol.

clear all;
close all;

%% ADD TENSORLAB TO PATH
path(path,['C:\GitLab\Tools\TensorLab20\TensorLab20']);

%% PARAMETERS
% geometric:
% H = 1; % height, in m
% L = 2; % length, in m
 thick = 1; % check this out to be sure !!!!!!!!!!!!!!

% prescribed pressure and pressure gradient, in m
% P = 0; % bottom and top 
% V = 0; % left and right

% material properties
rho     = 1.225; % [kg/m3]
c       = 343 ;  % [m/s]

% define the basis
b  = {'e1'; 'e2'};
ee = cartesianbasis2d(b{1}, b{2});
e1 = ee(1);
e2 = ee(2);


%% MANUALLY IMPORT COMSOL MESH 
% open txt file comsol_mesh_solid_rect.mphtxt and copy/paste the node 
% position and connectivity matrix.
% -------------------------------------------------------------------------
%  Comments on 'conn' matrix                                         STATUS
% -------------------------------------------------------------------------
% - swap column 3 and 4 of conn ( quadrilateral element);                OK
% - no need for swaping when triangular element;                         OK
% - add 1 to all nodes of 'conn'. In comsol it starts on 0, in matlab,1; OK
% - attention to units! Use SI.                                          OK
% - check if the mesh has at least 6 nodes per wavelength          ON-GOING


% load manually imported comsol mesh
load('es_solid_rect_MESH.mat');

% plot mesh
figure(1)
femplot(mesh_rect.x,mesh_rect.conn,'Color' , 'black', 'Nodes', 'off', ...
'NodeNumbers' , 'on', 'ElementNumbers', 'on');

% mesh parameters
m = size(mesh_rect.conn, 1); % number of elements in the mesh
n = size(mesh_rect.x, 1);    % number of nodes in the mesh 

%% INTEGRATION POINTS WITHIN MASTER ELEMENT

% coordinates of integration points 
% (in the coordinate system of the master element)
% and the weight factors
xi = [ -1/sqrt(3)*e1 - 1/sqrt(3)*e2
        1/sqrt(3)*e1 - 1/sqrt(3)*e2
        1/sqrt(3)*e1 + 1/sqrt(3)*e2
       -1/sqrt(3)*e1 + 1/sqrt(3)*e2 ];
w  = [ 1
       1
       1
       1 ];

%% ASSEMBLE SYSTEM MATRICES IN ELEMENT LOOP 
% following the notation in my notes
% A - [1/Pa]      equivalent "mass matrix", term that multiplies d^2/dt^2 p 
% B - [1/(kg/m3)] equivalent "stiffness matrix", term that multiplies  p 
% d - [m/s2]      RHS term "external forces"

A = zeros(n, n); % zero matrix of nxn scalars 
B = zeros(n, n); % zero matrix of nxn scalars

% copy to shorter variable names
conn = mesh_rect.conn;
x    = mesh_rect.x;

   for e = 1:m % loop over all elements

    iie = conn(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop
    Ae = zeros(4, 4);        % zero matrix of 4x4 scalars                            
    Be = zeros(4, 4);        % zero matrix of 4x4 scalars

    for k = 1:4 % loop over integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ 1/4*(1-xi1)*(1-xi2)
               1/4*(1+xi1)*(1-xi2)
               1/4*(1+xi1)*(1+xi2)
               1/4*(1-xi1)*(1+xi2) ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ -1/4*(1-xi2)*e1 - 1/4*(1-xi1)*e2
                      1/4*(1-xi2)*e1 - 1/4*(1+xi1)*e2
                      1/4*(1+xi2)*e1 + 1/4*(1+xi1)*e2
                     -1/4*(1+xi2)*e1 + 1/4*(1-xi1)*e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Ae = Ae + w(k) *(1/rho)*  Ne  * (1/c^2) * Ne' * thick * det(J);
        Be = Be + w(k) *(1/rho)*  dot(gradNe,gradNe') * thick * det(J);     
   
    end % end of interation point loop
   
% assembly
    A(iie, iie) = A(iie, iie) + Ae;
    B(iie, iie) = B(iie, iie) + Be;
end % end of element loop

%% APPLY BOUNDARY CONDITIONS BY PARTITIONING

% Manually select nodes by looking at the femplot
bottom = [1 3 6 10 14 18]; % nodes on the bottom boundary of the mesh
left   = [1 2 4 7]; % nodes on the left boundary of the mesh
right  = [18 21 23 24]; % nodes on the right boundary of the mesh
top    = [7 11 15 19 22 24]; % nodes on the top boundary of the mesh

% nodes where essential (Dirichlet) b.c. are applied p=0
iip = [left right];

% nodes where natural (Neumann) b.c. are applied ( free surface: 
% v=0). iit = [bottom top];

% the remaining nodes ( free nodes + Neumann BC v=0)
iif = setdiff(1:n, iip);

% mass matrix (tensor form)
App = A(iip, iip);
Afp = A(iif, iip);
Apf = A(iip, iif);  
Aff = A(iif, iif);

% stiffness matrix
Bpp = B(iip, iip);
Bfp = B(iif, iip);
Bpf = B(iip, iif);  
Bff = B(iif, iif);

%% VALUES OF THE PRESCRIBED DISPLACEMENTS
% prescribe pressure to zero (open)
pp  = [ zeros(length(left),1) ; zeros(length(right),1)]; 

% Imposed external acceleration dv/dt = zero inward the fluid domain along
% [bottom top] nodes, i.e., hard wall.
% Imposed external acceleration = zero along inner (free) nodes.
% Hence, the sum of [bottom top] nodes plus [inner] nodes results in
% [bottom top inner] which is equal to iff=[allnodes]-[left right]. 'Minus  
% sign' here should be interpreted in a set sense (set difference).
% RHS is therefore a vector of zeros on iif nodes.
df  = zeros(length(iif), 1);


%% EIGEN STUDY, PRINT EIGENFREQUENCIES

% define desired number of modes to compute
n_modes = 7;

% compute eigen modes V, and eigen values D with smallest absolute value
[V,D] = eigs(Bff,Aff,n_modes,"smallestabs");  
% compute frequencies omega
omega = sqrt(D);
freqs = diag(omega/(2*pi));
fprintf('Eigen frequencies:\n');
fprintf('%f Hz\n', freqs);

% %compute full matrix V and D
% [V_full,D_full] = eig(sKff,sMff); 
% omega_full = sqrt(D_full);
% freqs_full = diag(omega_full/(2*pi));
% fprintf('Eigen frequencies:\n');
% fprintf('%f Hz\n', sort(freqs_full, 'ascend'));

%% normalize eigen vectors with respect to mass matrix
% vec_norms = sqrt(diag(V_full'*Aff*V_full));
% V_full=V_full./vec_norms'; %divide each column of V by element of vec_norms 
% % safety check: compute again V'*Aff*V. It should be identity matrix.
% V=V_full;

%% ASSIGN SOLUTION TO DISPLACAMENT tensor-vector on each mode

% declare modal pressure all nodes for each mode
% p(nodes, modes)
p = zeros(n,size(V,2));

% free nodes, modal pressure shape
pf = V;
p(iif,:) = pf;

% essential nodes (dirichlet)
for mode=1:n_modes
    p(iip,size(V,2)) = pp;
end



%% PLOT MODE SHAPE (NON NORMALIZED)
mode=5;
figure(2)
clf
% plot the pressure solution
scatter(dot(x,e1),dot(x,e2),500,-p(:,mode),'filled')
% colorbar('Ticks',[-1,-.5,0,.5 1]);
colorbar;
caxis([-1 1])
colormap('jet');
axis equal
hold on
femplot(x, conn, 'Color', 'r');
title("Mode shape of frequency "  + num2str(freqs(mode)) + " Hz" );
hold off

