%% DESCRIPTION
%--------------------------------------------------------------------------
% 08/02/2022 Eigen study of fluid-solid rectangle
% R. Liupekevicius Carnielli
% the goal of this script is to compute a simple eigenstudy and validate it
% against comsol simulation. 
%
%
% based com "compression_solution.m" from ACCM and "es_solid_rect.m"
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  - free          = traction = 0;
%  - fixed         = displac. = 0;
%  - open          = p=0;
%  - f-s interface = continuity of normal velocity and stress;
%
%   (fixed)._____________(free)_____________. (fixed)
%          |                                |
%          |           solid(tag=3)         |
%(periodic)|                                | (periodic)
%          |________ (f-s interface)________|
%          |                                |
%          |                                |
%  (open)  |           fluid(tag=2)         |  (open)
%          |                                |
%          |                                |
%          |________ (f-s interface)________|
%          |                                |
%(periodic)|           solid(tag=1)         | (periodic)
%          |                                |
%  (fixed).|_____________(free)_____________|. (fixed)
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
% What are the low frequency modes ?
%
% Make sure you also have same element type in comsol.
%--------------------------------------------------------------------------
clear; close all;
tic
%% ADD TENSORLAB & SHARED FUNCTIONS & MESHES TO PATH
path(path,['C:\GitLab\Tools\TensorLab20\TensorLab20']);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\shared_functions"]);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\meshes"]);


%% PARAMETERS
% geometric:
H = 1; % height, in [m]
L = 2; % length, in [m]
thick = 1;  % thickness (out-of-plane direction), in [m]
% prescribed displacement, in [m]
U = 0;
% solid material properties
G        = 1000e6; % in [Pa]
kappa    = 3000e6; % in [Pa]
rhos     = 8000;   % in [kg/m3]

% solid material properties PU (IN TERMS OF E AND nu)
% rhos     = 1000;   % in [kg/m3]
% E        = 1e6;    % [Pa]
% nu       = 0.4;    % []    
% G        = E/2/(1+  nu);
% kappa    = E/3/(1-2*nu);

% fluid properties
rhof     = 1.225; % [kg/m3]
c        = 343 ;  % [m/s]
% pA       = 101325; [Pa]   % not used for order 1 equations.

% define the basis
b  = {'e1'; 'e2'};
ee = cartesianbasis2d(b{1}, b{2});
e1 = ee(1);
e2 = ee(2);


% determine material coefficient tensor
I   = identity(2, ee);
I4  = identity(4, ee);
I4S = 1/2 * (I4 + rtranspose(I4));
C   = kappa * I*I + 2*G * (I4S - 1/3*I*I);

%% MANUALLY IMPORT COMSOL MESH AS STRUCT WITH CELL ELEMENTS
%--------------------------------------------------------------------------
% open txt file comsol_mesh_solid_rect.mphtxt and copy/paste the node 
% position and connectivity matrix.
% -------------------------------------------------------------------------
%  Comments on 'conn' and 'tag' matrices for elements and edges      STATUS
% -------------------------------------------------------------------------
% - swap column 3 and 4 of conn ( quadrilateral element);                OK
% - no need for swaping when triangular element;                         OK
% - add 1 to all nodes of 'conn' (maybe 'tag' too). In comsol it starts
% in 0, whereas in matlab, it starts in 1;                               OK
% - attention to units! Use SI.                                          OK
% - check if the mesh has at least 6 nodes per wavelength          ON-GOING
%
%
%
% -------------------------------------------------------------------------
%  Comments on edge/element tags in comsol mesh file                                
% -------------------------------------------------------------------------
% - list of fluid-solid edge tag: 4,6
% This means that if an edge has tag 4 or 6, than the edge is at the fluid-
% solid interface.
% - list of element type (fluid or solid) after 'conn' matrix
%--------------------------------------------------------------------------

% load manually imported comsol mesh (use same mesh as in solid case)
load('es_fluidsolid_rect_MESH.mat');

%--------------------------------------------------------------------------
% convert to container.map object
% Recap:    Object Functions
% isKey	    Determine if Map object contains key
% keys	    Return keys of Map object
% length	Number of key-value pairs in Map object
% remove	Delete key-value pairs from Map object
% size   	Size of Map object
% values	Return values of Map object
%--------------------------------------------------------------------------


% copy to local variables (please double click on 'mesh' in workspace to
% see the meaning of each cell)
mesh = l;
x    = mesh.x{2,2};    % tensor
conn = mesh.elem{2,1}; % conn matrix of elements

% plot mesh
figure(1)
femplot(x,conn,'Color' , 'black', 'Nodes', 'off', ...
'NodeNumbers' , 'on', 'ElementNumbers', 'on');



% mesh parameters
m  = size(conn, 1); % number of elements in the mesh
n  = size(x, 1);    % number of nodes in the mesh 


% SOLID
    % solid elements
    % solid_elems = [find(mesh.elems('tag')==1);find(mesh.elems('tag')==3)].';
    solid_elems = union(find(mesh.elems('tag')==1),find(mesh.elems('tag')==3)).';
    % position vector of the solid nodes
    xs          = x(solid_elems);
    % conn solid
    conns       = conn(solid_elems,:);
    % solid nodes
    nodes_s     = unique(reshape(conns,1,[]));
    % number of solid elements in the mesh
    ms          = size(conns,1);
    % number of solid nodes
    ns          = length(nodes_s);


% FLUID
    % fluid elements
    fluid_elems = setdiff(1:m,solid_elems);
    % position vector of the fluid nodes
    xf          = x(fluid_elems);
    % conn fluid
    connf       = conn(fluid_elems,:);
    % fluid nodes
    nodes_f     = unique(reshape(connf,1,[]));
    % number of fluid elements in the mesh
    mf          = size(connf,1);
    % number of fluid nodes
    nf          = length(nodes_f);


%% INTERFACE NODES USING CONTAINER.MAP OBJECT AS DATA STRUCTURE
%--------------------------------------------------------------------------
% example on how to declair a map object of string keys with any value type
% map = containers.Map('KeyType', 'char', 'ValueType', 'any');

% redefine struct 'mesh' using container.map objects
% mesh.edges('conn')  is the connectivity matrix of edge elements. The
% indices of this vector represent the element number, i.g., if you want to
% access the second edge element
%--------------------------------------------------------------------------

% import edge tags from comsol 
tags_itr_edges=[4 6];
%(make 'for' loop below if list is too long)

% 'interface edges' are the indices of 'conn' matrix
itr_edges_tag4 = find(mesh.edges('tag')==4).';
itr_edges_tag6 = find(mesh.edges('tag')==6).';
itr_edges  = [itr_edges_tag4 itr_edges_tag6];
% these are edges(two nodes per edge) that form the interface.
% The edge number ( or edge 'element') is the index of 'conn'

% mesh.edge('conn')( interface_edges , :)
conn_edges = mesh.edges('conn'); %temporary assignment
conni_tag4 = conn_edges(itr_edges_tag4,:);
conni_tag6 = conn_edges(itr_edges_tag6,:);
conni      = conn_edges(itr_edges,:);
clear conn_edges;

% Get interface nodes (2D)
% - reshape conn (edges) matrix on a concatenated vector
% - keep 'unique' nodes of concatenated vector, i.e. remove duplicates
% - interface_nodes is therefore the index of elements' conn matrix
nodes_itr4     = unique(reshape( conni_tag4,1,[] ) ).';
nodes_itr6     = unique(reshape( conni_tag6,1,[] ) ).';
nodes_itr      = unique(reshape( conni     ,1,[] ) ).';



% POTENTIAL ALTERNATIVE
% nodes_itr =  intersect(nodes_s,nodes_f);

%%  4 INTEGRATION POINTS WITHIN MASTER ELEMENT
%--------------------------------------------------------------------------
% Notation from TensorLab
% Linear quadrilateral element:
%         4-------3
%         | x   x |
%         |       |
%         | x   x |
%         1-------2
%      then CONN = [1 2 3 4].
% coordinates of integration points x
% (in the coordinate system of the master element)
% and the weight factors
%--------------------------------------------------------------------------

xi = [ -1/sqrt(3)*e1 - 1/sqrt(3)*e2
        1/sqrt(3)*e1 - 1/sqrt(3)*e2
        1/sqrt(3)*e1 + 1/sqrt(3)*e2
       -1/sqrt(3)*e1 + 1/sqrt(3)*e2 ];
w  = [ 1
       1
       1
       1 ];

%% ASSEMBLE SYSTEM MATRICES IN A SOLID ELEMENT LOOP 
%--------------------------------------------------------------------------
% Knn and Mnn are only temporary matrix and they are not the matrices K and
% M from discretized form in the notes. Note their size is the total number
% of nodes which include the fluid nodes. It's only relevant property is
% that it preserves the node indexing.
%--------------------------------------------------------------------------

Knn = zeros(2, ee, n, n); % zero matrix of nxn 2nd order tensors in basis ee
Mnn = zeros(n, n);        % zero matrix of nxn scalars

   for e = 1:size(conns,1) % loop over all solid elements

    iie = conns(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop
    Ke = zeros(2, ee, 4, 4); % zero matrix of 4x4 2nd order tensors
                             % in basis ee.
    Me = zeros(4,4);         % zero matrix of 4x4 scalars

    for k = 1:4 % loop over integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ 1/4*(1-xi1)*(1-xi2)
               1/4*(1+xi1)*(1-xi2)
               1/4*(1+xi1)*(1+xi2)
               1/4*(1-xi1)*(1+xi2) ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ -1/4*(1-xi2)*e1 - 1/4*(1-xi1)*e2
                      1/4*(1-xi2)*e1 - 1/4*(1+xi1)*e2
                      1/4*(1+xi2)*e1 + 1/4*(1+xi1)*e2
                     -1/4*(1+xi2)*e1 + 1/4*(1-xi1)*e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Ke = Ke + w(k) *  dot(gradNe, C, gradNe') * thick * det(J);
        Me = Me + w(k) *         Ne *rhos*   Ne'  * thick * det(J);     
   
    end % end of interation point loop
   
   % assembly
    Knn(iie, iie) = Knn(iie, iie) + Ke;
    Mnn(iie, iie) = Mnn(iie, iie) + Me;
   end % end of element loop

% assignment actual mass and stiffness matrices
K=Knn(nodes_s,nodes_s);
M=Mnn(nodes_s,nodes_s);

%% ASSEMBLE SYSTEM MATRICES IN A FLUID ELEMENT LOOP 
%--------------------------------------------------------------------------
% following the notation in my notes
% A - [1/Pa]      equivalent "mass matrix", term that multiplies d^2/dt^2 p 
% B - [1/(kg/m3)] equivalent "stiffness matrix", term that multiplies  p 
% d - [m/s2]      RHS term "external forces"
%--------------------------------------------------------------------------

Ann = zeros(n, n); % zero matrix of nxn scalars 
Bnn = zeros(n, n); % zero matrix of nxn scalars


   for e = 1:size(connf,1)% loop over all fluid elements

    iie = connf(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop
    Ae = zeros(4, 4);        % zero matrix of 4x4 scalars                            
    Be = zeros(4, 4);        % zero matrix of 4x4 scalars

    for k = 1:4 % loop over integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ 1/4*(1-xi1)*(1-xi2)
               1/4*(1+xi1)*(1-xi2)
               1/4*(1+xi1)*(1+xi2)
               1/4*(1-xi1)*(1+xi2) ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ -1/4*(1-xi2)*e1 - 1/4*(1-xi1)*e2
                      1/4*(1-xi2)*e1 - 1/4*(1+xi1)*e2
                      1/4*(1+xi2)*e1 + 1/4*(1+xi1)*e2
                     -1/4*(1+xi2)*e1 + 1/4*(1-xi1)*e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Ae = Ae + w(k) *(1/rhof)*  Ne  * (1/c^2) * Ne' * thick * det(J);
        Be = Be + w(k) *(1/rhof)*  dot(gradNe,gradNe') * thick * det(J);     
   
    end % end of interation point loop
   
% assembly
    Ann(iie, iie) = Ann(iie, iie) + Ae;
    Bnn(iie, iie) = Bnn(iie, iie) + Be;
   end % end of fluid element loop

% define actual stiffness and mass matrices.
A=Ann(nodes_f,nodes_f);
B=Bnn(nodes_f,nodes_f);



%% INTEGRATION POINTS WITHIN MASTER EDGE (ELEMENT)
%--------------------------------------------------------------------------
% use here the same shape functions for both fluid and solid edges
% the edges have two nodes refered as 1 and 2 below.
%
%              solid              
%      1----x--------x-----2    
%              fluid             | (solid normal)   
%                                V 
% then CONN = [1 2].
% coordinates of integration points sketched as x.
%
% in the coordinate system of the master element and the weight factors
%--------------------------------------------------------------------------

xi = [ -1/sqrt(3)
        1/sqrt(3) ];
w  = [ 1
       1];

%% ASSEMBLE COUPLING MATRIX, LOOP ON INTERFACE EDGES 
%--------------------------------------------------------------------------
% use here the same shape functions for both fluid and solid elements
%--------------------------------------------------------------------------

Cnn = zeros(1, ee, n, n);        % zero matrix of nxn vectors


   for e = 1:size(conni,1) % loop over interface edges
    
    % edge e and its coordinates
    iie      = conni(e, :); % nodes of the current edge e
    xe       = x(iie);      % coordinates of these nodes
    
    % solid normal vector
    ve       = xe(2)-xe(1); % edge vector (of edge e)
    normal_s = ( dot(ve,e2)*e1 - dot(ve,e1)*e2)/...
                sqrt( dot(ve,e1)^2+dot(ve,e2)^2 ); % solid normal
%   normal_f = (-dot(ve,e2)*e1 + dot(ve,e1)*e2)/norm(ve); % fluid normal

    % compute element matrices in integration point loop
    Ce =  zeros(1, ee, 2, 2);        % 2x2 zero matrix of vectors

    for k = 1:2 % loop over integration points 
        
       % xi(k) is the coordinate of the current integr. point
       % column of the shape functions
        Ne = [ 1/2*(1-xi(k))
               1/2*(1+xi(k)) ];

       % elementary length ds=|dx|d(xi)       =  1/2*norm(ve)*d(xi) 
       % elementary area   dA=|dx|d(xi) thick =  1/2*norm(ve)*thick*d(xi)
       
    
       % element matrix 
        Ce = Ce + w(k) * (Ne * Ne') * normal_s * 1/2*norm(ve)* thick ;
        
         
   
    end % end of interation point loop
   
% assembly
    Cnn(iie, iie) = Cnn(iie, iie) + Ce;

   end % end of interface edge loop

C = Cnn(nodes_s,nodes_f);

% % check nonzero elements seing the following matrices
% % "primitive conversion" to scalar for a matter of checking nonzero compon.
% Cnncheck = dot(Cnn,e1).^2 + dot(Cnn,e2).^2; 
% Cnncheck(nodes_itr,nodes_itr);
% Cnncheck(setdiff(1:n,nodes_itr),setdiff(1:n,nodes_itr));

%% NODE SORTING
% Split the set of nodes 
% TODO: play with gmsh and automate the bc node sorting

% MANUALLY select nodes by looking at the femplot
% fluid
    % nodes on the left boundary of the mesh
    left_f    = [4 7 11 16 22]; 
    % nodes on the right boundary of the mesh
    right_f   = [123 129 134 138 141]; 
% solid
    % nodes on the bottom boundary of the mesh
    bottom_s  = [3 6 10 15 21 28 36 45 54 63 72 81 90 99];
    % nodes on the top boundary of the mesh
    top_s     = [46 55 64 73 82 91 100 109 117 124 130 135 139 142]; 
    % nodes on the left boundary of the mesh  (excluding corners)
    left_s    = [2 4 22 29]; 
    % nodes on the right boundary of the mesh  (excluding corners)
    right_s   = [116 123 141 143];
    % corner nodes
    corners_s = [1 108 144 37];
    

%% INDEX OF DISPLACEMENT/PRESSURE DOF IN NODES_S AND NODES_F
%--------------------------------------------------------------------------
% The matrices Mnn, Knn, Ann, Bnn, and Cnn have indices that correspond to
% the node number, i.g., line 1 are the terms related to node 1.
% The matrices M, K, A, B, and C, however, have indices that do not 
% correspond to the node number. Their indices correspond to the
% displacements degree of freedom or pressure degrees of freedom.
% Suppose, for example, you want to find the index in matrix M that
% correspond to the bottom-left corner displacement dof. This corner node
% is stored in the vector [corner], say first entry, corner(1). The index
% we are looking for, sometimes refered as 'new' index, is found using the
% following command:
% 
% find(nodes_s==corner(1)).
%
%--------------------------------------------------------------------------

% fluid
    % left nodes
    doff_l = get_ind(nodes_f, left_f   ); 
    % right nodes
    doff_r = get_ind(nodes_f, right_f  ); 
    % interface nodes
    doff_i = get_ind(nodes_f, nodes_itr);
% solid 
    % bottom nodes
    dofs_b = get_ind(nodes_s, bottom_s ); 
    % top nodes
    dofs_t = get_ind(nodes_s, top_s    ); 
    % left nodes
    dofs_l = get_ind(nodes_s, left_s   ); 
    % right nodes
    dofs_r = get_ind(nodes_s, right_s  ); 
    % interface nodes
    dofs_i = get_ind(nodes_s, nodes_itr);
    % corner nodes
    dofs_c = get_ind(nodes_s, corners_s);

%--------------------------------------------------------------------------
% TODO:afterwards, check which sets (above) of dof were actually used.
%--------------------------------------------------------------------------

%% ELIMINATING DEPENDENT NODES OF SOLID PHASE
%--------------------------------------------------------------------------
% constrain relation of periodicity for 'old indices'
% u([top_s right_s])=u([bottom_s left_s])
%
% constrain relation of periodicity for 'new indices'
% u([dofs_t dofs_r])=u([dofs_b dofs_l])
%--------------------------------------------------------------------------

    % dependent dofs of periodic bc: u(dofs_de)= u(dofs_in)
    dofs_de = [dofs_t dofs_r];   % [top right], excluding corners
    
    % independent nodes of periodic bc
    dofs_in = [dofs_b dofs_l];    %[bottom left], excluding corners
    

    % safety check
    if(length(dofs_in)~= length(dofs_de)) 
        error("The periodic bc must connect same number of nodes");
    end

% unconstrained nodes ( nodes not at the boundary + corners)
% [dofs_c] in [dofs_un] because they will be prescribed later on.
dofs_un = setdiff(1:ns, [dofs_in dofs_de]); 

%--------------------------------------------------------------------------
% reordering mass, stiffness, and coupling matrix components as
% [dofs_un dofs_in dofs_de]
%--------------------------------------------------------------------------

% mass scalar matrix & convert to tensor matrix
% M=  Muu Mui Mud   
%     Miu Mii Mid  
%     Mdu Mdi Mdd 
Muu=M(dofs_un,dofs_un)*I;Mui=M(dofs_un,dofs_in)*I;Mud=M(dofs_un,dofs_de)*I;  
Miu=M(dofs_in,dofs_un)*I;Mii=M(dofs_in,dofs_in)*I;Mid=M(dofs_in,dofs_de)*I;  
Mdu=M(dofs_de,dofs_un)*I;Mdi=M(dofs_de,dofs_in)*I;Mdd=M(dofs_de,dofs_de)*I;  
% stiffness tensor matrix
% K=  Kuu Kui Kud   
%     Kiu Kii Kid  
%     Kdu Kdi Kdd  
Kuu=K(dofs_un, dofs_un); Kui=K(dofs_un, dofs_in); Kud=K(dofs_un, dofs_de);  
Kiu=K(dofs_in, dofs_un); Kii=K(dofs_in, dofs_in); Kid=K(dofs_in, dofs_de);  
Kdu=K(dofs_de, dofs_un); Kdi=K(dofs_de, dofs_in); Kdd=K(dofs_de, dofs_de);  
% coupling tensor matrix
% C=  Cu    
%     Ci   
%     Cd  
Cu=C(dofs_un,:);
Ci=C(dofs_in,:);
Cd=C(dofs_de,:);
%--------------------------------------------------------------------------    
% projection of reordered K and M on [uu;ui], i.e., eliminate ud.
% projection: uu = [T] uu   or matlab notation  [uu;ui;ud]= [T] [uu;ui];
%             ui       ui
%             ud    
% the reduced matrices are
% mass tensor matrix      [M_] = [T]'[M][T]
% stiffness tensor matrix [K_] = [T]'[K][T]
% coupling tensor matrix  [C_] = [T]'[C]
% RHS tensor vector       [f_] = [T]'[f] (unknown on dirichlet nodes)
%                                        (kown on Neumman(=0)+free nodes)
%--------------------------------------------------------------------------

% projection matrix
T  = [                   eye(length([dofs_un dofs_in]))                  ;
       zeros(length(dofs_de),length(dofs_un))        eye(length(dofs_in))];

% reduced stiffness tensor matrix (dependent nodes eliminated)
K_ =  T'*[Kuu Kui Kud;
          Kiu Kii Kid;
          Kdu Kdi Kdd]*T;

% reduced mass tensor matrix (dependent nodes eliminated)
M_ =  T'*[Muu Mui Mud;
          Miu Mii Mid;
          Mdu Mdi Mdd]*T;
% reduced mass tensor matrix (dependent nodes eliminated)
C_ =  T'*[Cu;
          Ci;
          Cd];


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% question: projection with T' or T.' ? TODO: test both
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%--------------------------------------------------------------------------
% set of remaining nodes, this is the vector that contains the rules for
% changing the indices from before projection to after projection.
dofs_re = [dofs_un dofs_in];
ns_re   = length(dofs_re);
%--------------------------------------------------------------------------

%% PRESCRIBED AND FREE NODES SPLIT
%--------------------------------------------------------------------------
% INDEX OF DISPLACEMENT DOF IN NODES_S (ELIMINATED DEPENDENT DOFS)
% same index change procedure must be executed but only in the solid phase.

%|__NOTES_NOTATION_|____ DESCRIPTION_______________________________________
%|        p        | prescribed nodes in the solid phase
%|_________________|_______________________________________________________
%|                 |
%|        f        | free nodes in the solid phase                                                      
%|                 |
%|_________________|_______________________________________________________
%|                 |
%|        p'       | prescribed nodes in the fluid phase
%|_________________|_______________________________________________________
%|                 |
%|        f'       | free nodes in the fluid phase
%|_________________|_______________________________________________________
%--------------------------------------------------------------------------

% solid
p  = get_ind(dofs_re,dofs_c);
f  = setdiff(1:ns_re,p);

% fluid
pp = [doff_l doff_r];
fp = setdiff(1:nf,pp);

%% PARTITIONING

% solid
    % mass tensor matrix 
    % M =  Mpp Mpf   
    %      Mfp Mff 
    M_p_p   = M_(p , p ); M_f_p   = M_(f , p );
    M_p_f   = M_(p , f ); M_f_f   = M_(f , f );
    % stiffness tensor matrix
    % K =  Kpp Kpf   
    %      Kfp Kff 
    K_p_p   = K_(p , p ); K_f_p   = K_(f , p );
    K_p_f   = K_(p , f ); K_f_f   = K_(f , f );
    
% fluid 
    % 'mass' scalar matrix
    % A =  Ap'p' Ap'f'   
    %      Af'p' Af'f' 
    A_pp_pp = A (pp, pp); A_fp_pp = A (fp, pp);
    A_pp_fp = A (pp, fp); A_fp_fp = A (fp, fp);
    % 'stiffness' scalar matrix 
    % B =  Bp'p' Bp'f'   
    %      Bf'p' Bf'f' 
    B_pp_pp = B (pp, pp); B_fp_pp = B (fp, pp);
    B_pp_fp = B (pp, fp); B_fp_fp = B (fp, fp);

% coupling tensor matrix
    % C =  Cpp' Cpf'   
    %      Cfp' Cff' 
    C_p_pp  = C_(p , pp);C_p_fp  = C_(p , fp);
    C_f_pp  = C_(f , pp);C_f_fp  = C_(f , fp);

% zero matrices declaration
    O_f_fp     = zeros( size(ee,1)*length(f ) , length(fp)            );
    O_fp_f     = zeros( length(fp)            , size(ee,1)*length(f ) );

%% TRANFORM TENSOR MATRIX TO EQUIVALENT SCALAR MATRIX
%--------------------------------------------------------------------------
% TODO: rename, say M_f_f instead of sM_ff

% solid
    % mass matrix Mff
    sM_f_f      = tm2sm_v( M_f_f    , ee , 2 );
   
    % stiffness matrix Kff
    sK_f_f      = tm2sm_v( K_f_f    , ee , 2 );
    
    % coupling matrix C
    sC_f_fp     = tm2sm_v( C_f_fp   , ee , 1 );

%--------------------------------------------------------------------------
% fluid
    % 'mass' matrix A
    % (transformation not required)
    % 'stiffness' matrix B
    % (transformation not required)
    % coupling matrix C_f = - C^T
%--------------------------------------------------------------------------

%% ASSEMBLE NONSYMMETRIC SYSTEM OF EQUATIONS    
%--------------------------------------------------------------------------
% cond of the following matrices( with the term rhof*c2*) is lower
% % lamda matrix
% lamFF = [  sM_f_f                        O_f_fp  ;
%          -rhof*c^2*(sC_f_fp).'      rhof*c^2*A_fp_fp   ];  
% % mu matrix
% muFF  = [  sK_f_f              sC_f_fp ;
%             O_fp_f          rhof*c^2*B_fp_fp    ];  
%--------------------------------------------------------------------------

% lamda matrix
lamFF = [  sM_f_f           O_f_fp  ;
         -(sC_f_fp).'       A_fp_fp   ];  
% mu matrix
muFF  = [  sK_f_f              sC_f_fp ;
            O_fp_f              B_fp_fp    ];  

% norm_type='fro';
% fprintf('norm of M_f_f   = %f \n', norm(sM_f_f ,norm_type));
% fprintf('norm of K_f_f   = %f \n', norm(sK_f_f ,norm_type));
% fprintf('norm of A_fp_fp = %f \n', norm(A_fp_fp,norm_type));
% fprintf('norm of B_fp_fp = %f \n', norm(B_fp_fp,norm_type));
% fprintf('norm of C_f_fp  = %f \n', norm(sC_f_fp,norm_type));



%% CONDITIONING OF ASSEMBLED SYSTEM
fprintf('\n');
fprintf('cond(lamFF) = ');
disp(cond(lamFF));
fprintf('\n');
fprintf('cond( muFF) = ');
disp(cond(muFF));
fprintf('\n');

%% EIGEN STUDY, PRINT EIGENFREQUENCIES

% define desired number of modes to compute/display
n_modes = 20;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% compute all eigenvalues/eigenvectors %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
all=0;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if all==1
    % All eigenvalues %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % TODO: compute right/left eigen modes V,W, and eigen values D 
    method  = "qz";
    [phi,D,psi] = eig(muFF,lamFF,method); 
    
    
    % print frequencies f
    omega = sqrt(D);
    freqs = diag(omega/(2*pi)); % not necessarily 'ascend' order
    fprintf('Eigen frequencies:\n');
    fprintf('method: %s \n', method)
    [sfreqs,ifreqs]=sortrows(freqs); % sorted freqs 'ascend order'
    phi=phi(:,ifreqs); % rearange r-eigenvectors columns
    psi=psi(:,ifreqs); % rearange l-eigenvectors columns
    % W=W(:,ifreqs);
    fprintf('%f Hz\n', sfreqs(1:n_modes) );

else 
    % Subset of eigenvalues -> CHEAPER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % % eigs is recommended by matlab when sparse nonsymmetric matrices.
    % % compute eigen modes V, and eigen values D with smallest absolute
    % value right problem.
    [phi,Dr] = eigs(muFF ,lamFF ,n_modes,...
                   'smallestabs',...
                   'Display',1,...
                   'FailureTreatment','replacenan');
    % left problem
    [psi,Dl] = eigs(muFF',lamFF',n_modes,...
                   'smallestabs',...
                   'Display',1,...
                   'FailureTreatment','replacenan');
    
    % check whether the frequencies in Dr and Dl are enoughly equal. For
    % the current example they are equal up to the 5th digit after comma.
    
    % print frequencies f
    freqs = diag(sqrt(Dr)/(2*pi));
    fprintf('Eigen frequencies:\n');
    fprintf('%f Hz\n', freqs);
end

%% EIGEN MODES NORMALIZATION WITH RESPESCT TO THE LAMDA MATRIX
%--------------------------------------------------------------------------
% IMPORTANT NOTE: PHASE CORRECTION BEFORE NORMALIZATION 
% Compute round(psi'*lamFF*phi). It should be positive diag matrix - note 
% that this product should be approx positive diag because the the
% frequencies must be positive-. Since the right and left eigen modes were
% computed with different calls from the function eigs, they can
% potentially converge to a shape mode with phase difference of pi. In 
% this case, psi'*lamFF*phi ( or round(psi'*lamFF*phi) for getting rid 
% of almost zero off diag elements) may have some of the diagonal elements
% as -negative istead of +positive. To correct this effect it is necessary
% to multiply either phi or psi by the 'sign' matrix 'I'= psi'*lamFF*phi.
%--------------------------------------------------------------------------

% get diagonal
diag_psi_lam_phi   = diag(psi'*lamFF*phi);
% make it a vector with only the sign of each element
diag_psi_lam_phi   = diag_psi_lam_phi./abs(diag_psi_lam_phi);
% make 'identity' matrix with sign correction
I_phase_correction = diag(diag_psi_lam_phi);

% correct the sign of left eigen modes with respect to the right modes
psi = psi * I_phase_correction;

% ready to mass normalization, now, the product psi'*lamFF*phi has only
% postive diagonal compared to previous case
% get norm of each mode
vec_norms = sqrt(diag(psi'*lamFF*phi));
phi=phi./vec_norms'; %divide each column of phi by element of vec_norms 
psi=psi./vec_norms'; %divide each column of phi by element of vec_norms 

% CONSISTENCY CHECK: 
% Compute again psi'*lamFF*phi. It should be approx. the identity matrix.
% when eigs is used the phase problem doen't exist. It can be seen by
% computing psi(:,1:n_modes)'*lamFF*phi(:,1:n_modes). The diagonal is made
% out of positive integers.



%% ASSIGN DISP. SHAPE MODES TO TENSOR VECTOR
%--------------------------------------------------------------------------
% displacement 'ured' is followed is 'u' displacement reduced 'red' thanks
% to the elimitation of the dependend points due to periodic bc
% 'ured' stands for 
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% prescribed (reduced) displacements are zero
    % declare & assign
    ured_p = zeros(1, ee, length(p), size(phi,2) ); 
        

% free (reduced) displacements are the eigenvectors
    % declare
    ured_f = zeros(1, ee, length(f), size(phi,2));
    % assign
    % TODO: make the 'while' a function sv2tv() scalar vector to tensor vector
    i=1;
    while i<=length(f)
        ured_f(i,:) = phi(2*i-1,:)*e1+ phi(2*i,:)*e2;
        i=i+1;
    end

% total reduced displacements (indices ordered as [dofs_un dofs_in])
    % declare ured(unconstrained+corners+independent nodes, modes)
    ured = zeros(1, ee, length(p)+length(f), size(phi,2));
    % assign
    % assign prescribed nodes
    ured(p,:) = ured_p;
    % assign free nodes
    ured(f,:) = ured_f;
    % free unused variables
    clear ured_f; clear uref_p;
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% declare displacement tensor vectors (total # of dofs)
vu     = zeros(1, ee, ns, size(phi,2) );
vu_aux = zeros(1, ee, ns, size(phi,2) ); 
% assign. Note vu_aux is in order [dofs_un dofs_in dofs_de]
vu_aux = T*ured;
% assign to vu. Note vu is in nodes_s order
vu([dofs_un dofs_in dofs_de],:)  = vu_aux;
% free unused variables
clear vu_aux;

% prepare for plot
    % declare pressure scalar vector for plot
    vu_plot = zeros(1, ee, ns, size(phi,2) );
    % assign vp to vector for plot
    vu_plot(nodes_s,:) = vu;

%% ASSIGN PRESSURE SHAPE MODES TO SCALAR VECTOR vp

% extract pressure values from bc and eigen vectors
    % prescribed pressures (same for every mode)    
    p_pp     = zeros(length(pp), size(phi,2)); 
    
    % free pressures (different for each mode)
    p_fp = phi(end-length(fp)+1:end,:);

% declare pressure scalar vector
vp = zeros(nf,size(phi,2));

% assign to scalar vector solution of size
vp(pp,:)=p_pp;
vp(fp,:)=p_fp;

% prepare for plot
    % declare pressure scalar vector for plot
    vp_plot = zeros(n,size(phi,2));
    
    % assign vp to vector for plot
    vp_plot(nodes_f,:) = vp;


%% PLOT PRESSURE MODE SHAPE 

% % select 'mode'and plot settings
% mode=3;
% figure(2)
% clf
% % plot the pressure solution
% scatter(dot(x,e1),dot(x,e2),400,vp_plot(:,mode),'filled')
% % colorbar('Ticks',[-1,-.5,0,.5 1]);
% colorbar;
% pmax = max(abs(vp_plot(:,mode)));
% caxis([-pmax pmax]);
% colormap('jet');
% axis equal
% hold on
% femplot(x, conn, 'Color', 'k');
% title("Mode shape of frequency "  + num2str(freqs(mode)) + " Hz" );
% hold off

%% PLOT PRESSURE/DISP MODE SHAPE (RESCALED DISP.)
sfreqs=freqs; %sorted frequencies vector
mode=6;
figure(4)
clf
sign=1; % control modes shape phase ( 0 or 180 degrees)

% plot the pressure solution and displacement
scatter(dot(x,e1),dot(x,e2),400,sign*vp_plot(:,mode),'filled')
colorbar;
pmax = max(abs(vp_plot(:,mode)));
caxis([-pmax pmax]);
colormap('jet');
axis equal
hold on
vu_plot_n=sign*vu_plot(:,mode)/norm(norm(vu(:,mode))); % rescale
femplot(x+vu_plot_n, conn, 'Color', 'k');
title("Mode shape of frequency "  + num2str(sfreqs(mode)) + " Hz" );
hold off



toc

% % clear unused variables
% clear u_f; clear u_p; clear p_fp; clear p_pp;


% %% BUILD SOLUTION VECTORS
% % declare displacement tensor/scalar vector for each mode, using size n for
% % taking advantage of the indices.
%     % solid (vector of vector components)
%     vu = zeros(1, ee, n, size(phi,2));
%     % fluid (vector of scalar components)
%     sp = zeros(       n, size(phi,2));
% 
% % essential nodes (dirichlet)
% for mode=1:n_modes
%     vu(p ,mode) = u_p;
%     sp(pp,mode) = p_pp;
% end
% 
% % free nodes
% vu(f ,:) = u_f;
% sp(fp,:) = p_fp;
% 
% 
% %% PLOT PRESSURE MODE SHAPE 
% mode=1;
% figure(2)
% clf
% % plot the pressure solution
% scatter(dot(x,e1),dot(x,e2),400,sp(:,mode),'filled')
% % colorbar('Ticks',[-1,-.5,0,.5 1]);
% colorbar;
% pmax = max(abs(sp(:,mode)));
% caxis([-pmax pmax]);
% colormap('jet');
% axis equal
% hold on
% femplot(x, conn, 'Color', 'k');
% title("Mode shape of frequency "  + num2str(freqs(mode)) + " Hz" );
% hold off
% 
% %% PLOT DISPLACEMENT MODE SHAPE (RESCALED DISP.)
% mode=2;
% figure(3)
% clf
% vu_plot=vu(:,mode)/norm(norm(vu(:,mode)));
% % plot the displacement solution
% quiver(x, vu_plot, 0);
% axis equal
% hold on
% femplot(x+vu_plot, conn, 'Color', 'k');
% % femplot(vu_plot, conn, 'Color', 'k');
% title("Mode shape of frequency "  + num2str(freqs(mode)) + " Hz" );
% hold off
% 
% %% PLOT PRESSURE/DISP MODE SHAPE (RESCALED DISP.)
% sfreqs=freqs; %sorted frequencies vector
% mode=2;
% figure(4)
% clf
% sign=-1; % control modes shape phase ( 0 or 180 degrees)
% 
% % plot the pressure solution and displacement
% scatter(dot(x,e1),dot(x,e2),400,sign*sp(:,mode),'filled')
% colorbar;
% pmax = max(abs(sp(:,mode)));
% caxis([-pmax pmax]);
% colormap('jet');
% axis equal
% hold on
% vu_plot=sign*vu(:,mode)/norm(norm(vu(:,mode)));
% femplot(x+vu_plot, conn, 'Color', 'k');
% title("Mode shape of frequency "  + num2str(sfreqs(mode)) + " Hz" );
% hold off



