%% DESCRIPTION
% 09/01/2022 Eigen study of solid rectangle
% R. Liupekevicius Carnielli
%the goal of this script is to compute a simple eigenstudy and validate it
% against comsol simulation. 
%
%
% based com "compression_solution.m" from ACCM
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  - free = traction = 0;
%  - fixed= displac. = 0;
%
%          ___(free)__
%        /|           |\
% (fixed)/|           |\ (fixed)
%        /|           |\
%        /|___(free)__|\
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
% What are the low frequency modes of this structure under plane strain
% assumption?
%
% Make sure you also use same element type in comsol.

clear all;
close all;

%% ADD TENSORLAB & SHARED FUNCTIONS & MESHES TO PATH
path(path,['C:\GitLab\Tools\TensorLab20\TensorLab20']);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\shared_functions"]);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\meshes"]);


%% PARAMETERS
% geometric:
H = 1; % height, in m
L = 2; % length, in m
thick = 0.2;  % thickness (in out-of-plane direction), in m
% prescribed displacement, in m
U = 0;
% material properties
G     = 1000e6; % in [Pa]
kappa = 3000e6; % in [Pa]
rho   = 8000; % in [kg/m3]

% define the basis
b  = {'e1'; 'e2'};
ee = cartesianbasis2d(b{1}, b{2});
e1 = ee(1);
e2 = ee(2);

% determine material coefficient tensor
I   = identity(2, ee);
I4  = identity(4, ee);
I4S = 1/2 * (I4 + rtranspose(I4));
C   = kappa * I*I + 2*G * (I4S - 1/3*I*I);

%% MANUALLY IMPORT COMSOL MESH 
% open txt file comsol_mesh_solid_rect.mphtxt and copy/paste the node 
% position and connectivity matrix.
% -------------------------------------------------------------------------
%  Comments on 'conn' matrix                                         STATUS
% -------------------------------------------------------------------------
% - swap column 3 and 4 of conn ( quadrilateral element);                OK
% - no need for swaping when triangular element;                         OK
% - add 1 to all nodes of 'conn'. In comsol it starts on 0, in matlab,1; OK
% - attention to units! Use SI.                                          OK
% - check if the mesh has at least 6 nodes per wavelength          ON-GOING

% load manually imported comsol mesh
load('es_solid_rect_MESH.mat');

% plot mesh
figure(1)
femplot(mesh_rect.x,mesh_rect.conn,'Color' , 'black', 'Nodes', 'off', ...
'NodeNumbers' , 'on', 'ElementNumbers', 'on');

% mesh parameters
m = size(mesh_rect.conn, 1); % number of elements in the mesh
n = size(mesh_rect.x, 1);    % number of nodes in the mesh 

%% INTEGRATION POINTS WITHIN MASTER ELEMENT

% coordinates of integration points 
% (in the coordinate system of the master element)
% and the weight factors
xi = [ -1/sqrt(3)*e1 - 1/sqrt(3)*e2
        1/sqrt(3)*e1 - 1/sqrt(3)*e2
        1/sqrt(3)*e1 + 1/sqrt(3)*e2
       -1/sqrt(3)*e1 + 1/sqrt(3)*e2 ];
w  = [ 1
       1
       1
       1 ];

%% ASSEMBLE SYSTEM MATRICES IN ELEMENT LOOP 
K = zeros(2, ee, n, n); % zero matrix of nxn 2nd order tensors in basis ee
M = zeros(n, n);        % zero matrix of nxn scalars

% copy to shorter variable names
conn = mesh_rect.conn;
x    = mesh_rect.x;

   for e = 1:m % loop over all elements

    iie = conn(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop
    Ke = zeros(2, ee, 4, 4); % zero matrix of 4x4 2nd order tensors
                             % in basis ee.
    Me = zeros(4,4);         % zero matrix of 4x4 scalars

    for k = 1:4 % loop over integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ 1/4*(1-xi1)*(1-xi2)
               1/4*(1+xi1)*(1-xi2)
               1/4*(1+xi1)*(1+xi2)
               1/4*(1-xi1)*(1+xi2) ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ -1/4*(1-xi2)*e1 - 1/4*(1-xi1)*e2
                      1/4*(1-xi2)*e1 - 1/4*(1+xi1)*e2
                      1/4*(1+xi2)*e1 + 1/4*(1+xi1)*e2
                     -1/4*(1+xi2)*e1 + 1/4*(1-xi1)*e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Ke = Ke + w(k) * dot(gradNe, C, gradNe') * thick * det(J);
        Me = Me + w(k) *         Ne *rho*   Ne'  * thick * det(J);     
   
    end % end of interation point loop
   
% assembly
    K(iie, iie) = K(iie, iie) + Ke;
    M(iie, iie) = M(iie, iie) + Me;
end % end of element loop

%% APPLY BOUNDARY CONDITIONS BY PARTITIONING

% Manually select nodes by looking at the femplot
bottom = [1 3 6 10 14 18]; % nodes on the bottom boundary of the mesh
left   = [1 2 4 7]; % nodes on the left boundary of the mesh
right  = [18 21 23 24]; % nodes on the right boundary of the mesh
top    = [7 11 15 19 22 24]; % nodes on the top boundary of the mesh

% nodes where essential (Dirichlet) b.c. are applied
iiu = [left right];

% nodes where natural (Neumann) b.c. are applied ( free surface: 
% traction=0). iit = [bottom top];

% the remaining nodes ( free nodes + Neumann BC)
iif = setdiff(1:n, iiu);

% stiffness matrix
Kuu = K(iiu, iiu);
Kfu = K(iif, iiu);
Kuf = K(iiu, iif);  
Kff = K(iif, iif);

% mass matrix (tensor form)
Muu = M(iiu, iiu)*I;
Mfu = M(iif, iiu)*I;
Muf = M(iiu, iif)*I;  
Mff = M(iif, iif)*I;

%% VALUES OF THE PRESCRIBED DISPLACEMENTS
% prescribe to zero (clamped)
uu  = [ zeros(1, ee, length(left),1) ; zeros(1, ee, length(right),1)]; 

% Note: In this particlar case the traction is zero on bottom and top
% boundaries therefore ff is a zero vector for all iif nodes. If other
% bc is imposed then a different force should be assigned to 
% [bottom top] nodes.
ff  = zeros(1, ee, length(iif), 1);


%% TRANSFORM tensor-matrix into scalar-matrix

% TRANSFORMS stiffness tensor-matrix into array (its scalar matrix)
% make the code block a function with input: K, ee ( i couldn't because
% aparently it is not possible to pass a tensor to a function.
% spatial dimension
dim  = size(ee, 1);
switch dim
    case 2
        % number of free dofs
        nf   = dim * size(Kff,1);
        % define scalar matrix equivalent of Kff
        sKff = zeros(nf,nf);        
        for i=1:size(Kff,1) %loop over matrix tensor components
           for j=1:size(Kff,1)
               %submatrix
               aux  = [dot(e1,Kff(i,j),e1) dot(e1,Kff(i,j),e2);
                       dot(e2,Kff(i,j),e1) dot(e2,Kff(i,j),e2)];  
               % indices in the scalar matrix
               si=1+dim*(i-1):dim*i;
               sj=1+dim*(j-1):dim*j;
               sKff(si,sj) = sKff(si,sj) + aux;
           end
        end
    case 3
        % number of free dofs
        nf   = dim * size(Kff,1);
        % define scalar matrix equivalent of Kff
        sKff = zeros(nf,nf);
        for i=1:size(Kff,1) %loop over matrix tensor components
           for j=1:size(Kff,1)
               %submatrix
         aux=[dot(e1,Kff(i,j),e1) dot(e1,Kff(i,j),e2) dot(e1,Kff(i,j),e3);
              dot(e2,Kff(i,j),e1) dot(e2,Kff(i,j),e2) dot(e2,Kff(i,j),e3);
              dot(e3,Kff(i,j),e1) dot(e3,Kff(i,j),e2) dot(e3,Kff(i,j),e3)];
               % indices in the scalar matrix
               si=1+dim*(i-1):dim*i;
               sj=1+dim*(j-1):dim*j;
               sKff(si,sj) = sKff(si,sj) + aux;
           end
        end
end
% TRANSFORMS mass tensor-matrix into array (its equivalent scalar matrix)
% spatial dimension
dim  = size(ee, 1);
switch dim
    case 2
        % number of free dofs
        nf   = dim * size(Mff,1);
        % define scalar matrix equivalent of Kff
        sMff = zeros(nf,nf);
        for i=1:size(Mff,1)%loop over matrix tensor components
           for j=1:size(Mff,1)
           aux  = [dot(e1,Mff(i,j),e1) dot(e1,Mff(i,j),e2);
                   dot(e2,Mff(i,j),e1) dot(e2,Mff(i,j),e2)];  
           % indices in the scalar matrix
           si=1+dim*(i-1):dim*i;
           sj=1+dim*(j-1):dim*j;
           sMff(si,sj) = sMff(si,sj) + aux;
           end
        end
    case 3
        % number of free dofs
        nf   = dim * size(Mff,1);
        % define scalar matrix equivalent of Kff
        sMff = zeros(nf,nf);
        for i=1:size(Mff,1)%loop over matrix tensor components
           for j=1:size(Mff,1)
      aux  = [dot(e1,Mff(i,j),e1) dot(e1,Mff(i,j),e2) dot(e1,Mff(i,j),e3);
              dot(e2,Mff(i,j),e1) dot(e2,Mff(i,j),e2) dot(e2,Mff(i,j),e3);
              dot(e3,Mff(i,j),e1) dot(e3,Mff(i,j),e2) dot(e3,Mff(i,j),e3)];
           % indices in the scalar matrix
           si=1+dim*(i-1):dim*i;
           sj=1+dim*(j-1):dim*j;
           sMff(si,sj) = sMff(si,sj) + aux;
           end
        end
end

%% EIGEN STUDY, PRINT EIGENFREQUENCIES

% define desired number of modes to compute
n_modes = 7;

% % compute eigen modes V, and eigen values D with smallest absolute value
% [V,D] = eigs(sKff,sMff,n_modes,"smallestabs");  
% % compute frequencies omega
% omega = sqrt(D);
% freqs = diag(omega/(2*pi));
% fprintf('Eigen frequencies:\n');
% fprintf('%f Hz\n', freqs);


% balance
[V,D] = eigs(sKff,sMff,n_modes,"smallestabs");  
% compute frequencies omega
omega = sqrt(D);
freqs = diag(omega/(2*pi));
fprintf('Eigen frequencies:\n');
fprintf('%f Hz\n', freqs);


% %compute full matrix V and D
% [V_full,D_full] = eig(sKff,sMff); 
% omega_full = sqrt(D_full);
% freqs_full = diag(omega_full/(2*pi));
% fprintf('Eigen frequencies:\n');
% fprintf('%f Hz\n', sort(freqs_full, 'ascend'));

%% normalize eigen vectors with respect to mass matrix
% vec_norms = sqrt(diag(V_full'*sMff*V_full));
% V_full=V_full./vec_norms'; %divide each column of V by element of vec_norms 
% % safety check: compute again V'*sMff*V. It should be identity matrix.
% V=V_full;
%% TRANSFORM SOLUTION deformation mode from scalar-vector to tensor-vector
% uf(nodes, modes)
% declare displacement matrix tensor of free displacement (shape modes)
uf = zeros(1, ee, size(V,1)/dim,  size(V,2));

i=1;
while i<=size(V,1)/dim
    uf(i,:) = V(2*i-1,:)*e1+ V(2*i,:)*e2;
    i=i+1;
end

%% ASSIGN SOLUTION TO DISPLACAMENT tensor-vector on each mode

% declare displacement tensor-vector for each mode
u = zeros(1, ee, n, size(V,2));

% essential nodes (dirichlet)
for mode=1:n_modes
    u(iiu,mode) = uu;
end

% free nodes
u(iif,:) = uf;


%% PLOT MODE SHAPE (NON NORMALIZED)
mode=1;
figure(2)
clf
% plot the displacement solution
quiver(x, u(:,mode), 0);
axis equal
hold on
femplot(x+u(:,mode), conn, 'Color', 'r');
title("Mode shape of frequency "  + num2str(freqs(mode)) + " Hz" );
hold off

 %% PLOT MODE SHAPE (NORMALIZED)
% mode=23;
% figure(2)
% clf
% % plot the displacement solution
% quiver(x, u(:,mode), 0);
% axis equal
% hold on
% femplot(x+u(:,mode), conn, 'Color', 'r');
% title("Mode shape of frequency "  + num2str(freqs_full(mode)) + " Hz" );
% hold off
