%% DESCRIPTION
% 03/03/2022 Eigen study of solid rectangle (quadratic mesh)
% R. Liupekevicius Carnielli
% the goal of this script is to compute a simple eigenstudy and validate it
% against comsol simulation. 
%
%
% based com "compression_solution.m" from ACCM
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  - periodic: disp.(source) = disp.(destination);
%  - fixed   : disp. = 0;
%
%     (fixed).___(periodic)__.(fixed)
%            |               |
% (periodic) |               | (periodic)
%            |               |
%            |___(periodic)__|
%     (fixed)'               '(fixed)
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
% What are the low frequency modes of this structure under plane strain
% assumption?
%
% Make sure you also use same element type in comsol.

clear all;
close all;

%% ADD TENSORLAB & SHARED FUNCTIONS & MESHES TO PATH
path(path,['C:\GitLab\Tools\TensorLab20\TensorLab20']);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\shared_functions"]);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\meshes"]);

%% PARAMETERS
% geometric:
H = 1; % height, in m
L = 2; % length, in m
thick = 0.2;  % thickness (in out-of-plane direction), in m
% prescribed displacement, in m
U = 0;
% material properties
G     = 1000e6; % in [Pa]
kappa = 3000e6; % in [Pa]
rhos  = 8000; % in [kg/m3]

% define the basis
b  = {'e1'; 'e2'};
ee = cartesianbasis2d(b{1}, b{2});
e1 = ee(1);
e2 = ee(2);

% determine material coefficient tensor
I   = identity(2, ee);
I4  = identity(4, ee);
I4S = 1/2 * (I4 + rtranspose(I4));
C   = kappa * I*I + 2*G * (I4S - 1/3*I*I);

%% MANUALLY IMPORT COMSOL MESH 
% open txt file comsol_mesh_solid_rect.mphtxt and copy/paste the node 
% position and connectivity matrix.
% -------------------------------------------------------------------------
%  Comments on 'conn' matrix and comsol mesh file                    STATUS
% -------------------------------------------------------------------------
% - swap column 3 and 4 of conn (quadrilateral) to set tensorlab format; OK
% - no need for swaping when triangular element;                         OK
% - add 1 to all nodes of 'conn'. In comsol it starts on 0, in matlab,1; OK
% - attention to units! Use SI.                                          OK
% - check if the mesh has at least 6 nodes per wavelength          ON-GOING
%--------------------------------------------------------------------------

% load manually imported comsol mesh
load('solid_rect_MESH.mat');

% transform to quadratic element
q=mesh2d_lin2qua(l);


% copy to local variables (please double click on 'mesh' in workspace to
% see the meaning of each cell)
x    = q.x{2,2};    % tensor
conn = q.elem{2,1}; % conn matrix of elements

% % plot mesh
% figure(1)
% femplot(x,conn,'Color' , 'black', 'Nodes', 'off', ...
% 'NodeNumbers' , 'on', 'ElementNumbers', 'on');

% mesh parameters
m = size(conn, 1); % number of elements in the mesh
n = size(x, 1);    % number of nodes in the mesh 

%%  4 INTEGRATION POINTS WITHIN MASTER ELEMENT
%--------------------------------------------------------------------------
% Notation from TensorLab
% quadratic quadrilateral element:
%         4----7----3
%         |         |
%         8         6
%         |         |
%         1----5----2
%      then CONN = [1 2 3 4 5 6 7 8].
%
% The coordinates of integration points x (in the coordinate system of the
% master element). Exact integration up to polynomial of order 5.
%          --------- 
%         | x  x  x |
%         | x  x  x |
%         | x  x  x |
%          ---------
% and the weight factors
%--------------------------------------------------------------------------

xi = [ -sqrt(3/5) *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 + sqrt(3/5) *e2
       -sqrt(3/5) *e1 + sqrt(3/5) *e2
                0 *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 +         0 *e2       
                0 *e1 + sqrt(3/5) *e2                
       -sqrt(3/5) *e1 +         0 *e2
                0 *e1 +         0 *e2   ];
w  = [ 25/81
       25/81
       25/81
       25/81
       40/81
       40/81
       40/81
       40/81
       64/81     ];

%% ASSEMBLE SYSTEM MATRICES IN A SOLID ELEMENT LOOP 
%--------------------------------------------------------------------------
% Knn and Mnn are only temporary matrix and they are not the matrices K and
% M from discretized form in the notes. Note their size is the total number
% of nodes which include the fluid nodes. It's only relevant property is
% that it preserves the node indexing.
%--------------------------------------------------------------------------

K = zeros(2, ee, n, n); % zero matrix of nxn 2nd order tensors in basis ee
M = zeros(n, n);        % zero matrix of nxn scalars

   for e = 1:size(conn,1) % loop over all solid elements

    iie = conn(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop
    Ke = zeros(2, ee, 8, 8); % zero matrix of 8x8 2nd order tensors
                             % in basis ee.
    Me = zeros(8,8);         % zero matrix of 8x8 scalars

    for k = 1:length(w) % loop over 9 integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ -1/4*(1-xi1  )*(1-xi2  )*(1+xi1+xi2)
               -1/4*(1+xi1  )*(1-xi2  )*(1-xi1+xi2)
               -1/4*(1+xi1  )*(1+xi2  )*(1-xi1-xi2)
               -1/4*(1-xi1  )*(1+xi2  )*(1+xi1-xi2)
                1/2*(1-xi1^2)*(1-xi2  )
                1/2*(1+xi1  )*(1-xi2^2)
                1/2*(1-xi1^2)*(1+xi2  )
                1/2*(1-xi1  )*(1-xi2^2)           ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ - 1/4 * (-1 + xi2   )* (2*xi1 +   xi2) *e1 ...
                     - 1/4 * (-1 + xi1   )* (xi1   + 2*xi2) *e2
                     - 1/4 * (2*xi1 - xi2)* (-1    + xi2  ) *e1 ...
                     - 1/4 * (1 + xi1    )* (xi1   - 2*xi2) *e2
                       1/4 * (1 + xi2    )* (2*xi1 + xi2  ) *e1 ...
                     + 1/4 * (1 + xi1    )* (xi1   + 2*xi2) *e2
                       1/4 * (2*xi1 - xi2)* (1     + xi2  ) *e1 ...
                     + 1/4 * (-1 + xi1   )* (xi1 - 2*xi2  ) *e2
                                      xi1 * (-1 + xi2     ) *e1 ...
                                    + 1/2 * (-1 + xi1^2   ) *e2
                                      1/2 * ( 1 - xi2^2   ) *e1 ...
                                    - xi2 * ( 1 + xi1     ) *e2
                                    - xi1 * ( 1 + xi2     ) *e1 ...
                                    + 1/2 * ( 1 - xi1^2   ) *e2
                                      1/2 * (-1 + xi2^2   ) *e1 ...
                                    + xi2 * (-1 + xi1     ) *e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Ke = Ke + w(k) *  dot(gradNe, C, gradNe') * thick * det(J);
        Me = Me + w(k) *         Ne *rhos*   Ne'  * thick * det(J);     
   
    end % end of interation point loop
   
   % assembly
    K(iie, iie) = K(iie, iie) + Ke;
    M(iie, iie) = M(iie, iie) + Me;
   end % end of element loop



%% NODE SORTING

%spatial dimension
dim = size(ee,1);

% Manually select nodes by looking at the femplot
% nodes on the top boundary WITHOUT corners
[top,~] = find(q.x{2,1}(:,2)==1);
top = top(2:end-1)';
% top    = [25    28    30    35    39    43    48    53    58];

% nodes on the bottom boundary WITHOUT corners
bottom = [3     6    10    16    23    31    38    44    50];   
 % nodes on the left WITHOUT corners
left   = [2     4     8    12    18 ];         
% nodes on the right WITHOUT corners
right  = [56    57    59    60    61];      
% corner nodes on the boundary of the mesh
corners= [1 55 62 24];        

%% ELIMINATING DEPENDENT NODES

% constrain relation of periodicity u([top right])=u([bottom left])
    % dependent nodes of periodic bc: u(iid)= u(iii)
    iid = [top right];   % excluding corners
    
    % independent nodes of periodic bc
    iii = [bottom left]; % excluding corners
    

    % safety check
    if(length(iii)~= length(iid)) 
        error("The periodic bc must connect same number of nodes");
    end

% unconstrained nodes ( nodes not at the boundary + corners)
iiu = setdiff(1:n, [iii iid]); % [corners] in [iiu] because they
                                    % will be prescribed later on.

% reordering mass and stiffness elements as [iiu iii iid]
    % stiffness tensor matrix
    % K_reordered =  Kuu Kui Kud   
    %                Kiu Kii Kid  
    %                Kdu Kdi Kdd  
    Kuu = K(iiu, iiu); Kui = K(iiu, iii); Kud = K(iiu, iid);  
    Kiu = K(iii, iiu); Kii = K(iii, iii); Kid = K(iii, iid);  
    Kdu = K(iid, iiu); Kdi = K(iid, iii); Kdd = K(iid, iid);  
    % mass scalar matrix & convert to tensor matrix
    % M_reordered = Muu Mui Mud   
    %               Miu Mii Mid  
    %               Mdu Mdi Mdd 
    Muu = M(iiu, iiu)*I; Mui = M(iiu, iii)*I; Mud = M(iiu, iid)*I;  
    Miu = M(iii, iiu)*I; Mii = M(iii, iii)*I; Mid = M(iii, iid)*I;  
    Mdu = M(iid, iiu)*I; Mdi = M(iid, iii)*I; Mdd = M(iid, iid)*I;  

%--------------------------------------------------------------------------    
% projection of reordered K and M on [uu;ui], i.e., eliminate ud.
% projection: uu = [T] uu   or matlab notation  [uu;ui;ud]= [T] [uu;ui];
%             ui       ui
%             ud    
% the reduced matrices are
% mass tensor matrix      [M_] = [T]'[M][T]
% stiffness tensor matrix [K_] = [T]'[K][T]
% RHS tensor vector       [f_] = [T]'[f] (unknown on dirichlet nodes)
%                                        (kown on Neumman(=0)+free nodes)
%--------------------------------------------------------------------------

% projection matrix
T  = [                       eye(length([iiu iii]))                 ;
       zeros(length(iid),length(iiu))               eye(length(iii)) ];

% reduced stiffness tensor matrix
K_ =  T'*[Kuu Kui Kud;
          Kiu Kii Kid;
          Kdu Kdi Kdd]*T;

% reduced mass tensor matrix
M_ =  T'*[Muu Mui Mud;
          Miu Mii Mid;
          Mdu Mdi Mdd]*T;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% question: projection with T' or T.' ? TODO: test both
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% set of remaining nodes
iin = [iiu iii];

%% APPLY PERIODIC BOUNDARY CONDITIONS BY
%% PARTITIONING PRESCRIBED AND FREE NODES OF THE REDUCED MATRICES K_ AND M_
%--------------------------------------------------------------------------
% The indices in matrix K/M correspond to the node number, however the
% indices in matrices K_ and M_ DO NOT. It means that the vector iin 
% converts the node number to its new indices in matrix K_ and M_.
% Use the function find() to find the new index in matrix K_/M_. For
% example if you want to find the new index of node 18, use the following:
%
% find(iin==18);
%
% or if you want to find the new indices of the corners
% 
% forloop i: find(iin==corner(i))
%--------------------------------------------------------------------------

% old indices of prescribed nodes
iop = corners; % index/old/prescribed

% new indices of prescribed nodes
inp = [];      % index/new/prescribed
for i=1:length(iop)
inp = [inp find(iin==iop(i))];
end

% old indices of free nodes
iof = setdiff(iin,corners); % index/old/free

% new indices of free nodes
inf=[];                     % index/new/free
for i=1:length(iof)
inf = [inf find(iin==iof(i))];
end

% stiffness tensor matrix
Kpp = K_(inp, inp); Kpf = K_(inp, inf);
Kfp = K_(inf, inp); Kff = K_(inf, inf);

% mass tensor matrix
Mpp = M_(inp, inp); Mpf = M_(inp, inf);
Mfp = M_(inf, inp); Mff = M_(inf, inf);



%% TRANSFORM tensor-matrix into scalar-matrix
%--------------------------------------------------------------------------
% TRANSFORMS stiffness tensor-matrix into array (its scalar matrix equiv.)
% make the code block a function with input: K, ee? ( i couldn't because
% aparently it is not possible to pass a tensor to a function. Investigate!
%--------------------------------------------------------------------------

% mass matrix Mff
sMff      = tm2sm( Mff    , ee , 2 );

% stiffness matrix Kff
sKff      = tm2sm( Kff    , ee , 2 );

%% EIGEN STUDY, PRINT EIGENFREQUENCIES

% define desired number of modes to compute
n_modes = 7;

% compute eigen modes V, and eigen values D with smallest absolute value,
% i.e., the lower frequencies.
[V,D] = eigs(sKff,sMff,n_modes,"smallestabs");  

% compute frequencies omega
omega = sqrt(D);
freqs = diag(omega/(2*pi));
fprintf('Eigen frequencies:\n');
fprintf('%f Hz\n', freqs);

% %compute full matrix V and D
% [V_full,D_full] = eig(sKff,sMff); 
% omega_full = sqrt(D_full);
% freqs_full = diag(omega_full/(2*pi));
% fprintf('Eigen frequencies:\n');
% fprintf('%f Hz\n', sort(freqs_full, 'ascend'));

%% EIGEN VECTOR NORMALIZATION WITH RESPECT TO MASS MATRIX
% normalize eigen vectors with respect to mass matrix
vec_norms = sqrt(diag(V'*sMff*V));
V         = V./vec_norms'; %divide each column of V by element of vec_norms 
% safety check: compute again V'*sMff*V. It should be identity matrix.


%% APPLY DIRICHLET BOUNDARY CONDITION

% prescribe to zero (clamped)
ured_p  = [ zeros(1, ee, length(inp),1)]; 

%% TRANSFORM REDUCED DEFORMATION SHAPE FROM SCALAR VECTOR TO TENSOR VECTOR
% uf(free nodes, modes)
% declare displacement matrix tensor of free displacements (shape modes)
ured_f = zeros(1, ee, size(V,1)/dim,  size(V,2));

i=1;
while i<=size(V,1)/dim
    ured_f(i,:) = V(2*i-1,:)*e1+ V(2*i,:)*e2;
    i=i+1;
end

%% ASSIGN REDUCED SOLUTION TO DISPLACAMENT TENSOR VECTOR ured
% declare displacement tensor-vector for each mode
ured = zeros(1, ee, length(iin), size(V,2));

% prescribed nodes (dirichlet)
for mode=1:n_modes
    ured(inp,mode) = ured_p;
end

% free nodes
ured(inf,:) = ured_f;

%% PLOT DISPLACEMENT TENSOR VECTOR
% projection: [u]= [T] [ured] or [uu;ui;ud]= [T] [uu;ui];
% declare tensor vectors
uaux = zeros(1, ee, n, size(V,2)); % auxiliary tensor vector
u    = zeros(1, ee, n, size(V,2)); % displacement tensor vector

% assign components to displacement, current order is u([iiu iii iid])
uaux= T*ured;

% back to the initial node order u(1:n)
u([iiu iii iid],:) = uaux(:,:);

% PLOT MODE SHAPE (RESCALED)
mode=5;
u_plot=u(:,mode)/norm(norm(u(:,mode)));

figure(2)
clf
% plot the displacement solution
quiver(x, u_plot, 0);
axis equal
hold on
femplot(x+u_plot, conn, 'Color', 'r');
title("Mode shape of frequency "  + num2str(freqs(mode)) + " Hz" );
hold off

