%% DESCRIPTION
% 24/02/2022 Stationary solution study of fluid rectangle
% R. Liupekevicius Carnielli 
%
%
% based com "compression_solution.m" from ACCM
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  - pressure = p1 (constant in time)
%  - pressure = p2 (constant in time)
%  - hard: velocity = 0 ( or pressure gradient =0)
%
%          ___(hard)__
%        /|           |\
%  (p1  )/|           |\ (p2)
%        /|           |\
%        /|___(hard)__|\
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
% What is the pressure field?

clear all;
close all;

%% ADD TENSORLAB & SHARED FUNCTIONS & MESHES TO PATH
path(path,['C:\GitLab\Tools\TensorLab20\TensorLab20']);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\shared_functions"]);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\meshes"]);


%% PARAMETERS
% geometric:
% H = 1; % height, in m
% L = 2; % length, in m
 thick = 1; % check this out to be sure !!!!!!!!!!!!!!

% prescribed pressure and pressure gradient, in m
% P = 0; % bottom and top 
% V = 0; % left and right

% material properties
rho     = 1.225; % [kg/m3]
c       = 343 ;  % [m/s]

% define the basis
b  = {'e1'; 'e2'};
ee = cartesianbasis2d(b{1}, b{2});
e1 = ee(1);
e2 = ee(2);

%% BOUNDARY CONDITIONS
PL = 1;   %[Pa] at the left boundary
PR = 100; %[Pa] at the right boundary


%% MANUALLY IMPORT COMSOL MESH 
% open txt file comsol_mesh_solid_rect.mphtxt and copy/paste the node 
% position and connectivity matrix.
% -------------------------------------------------------------------------
%  Comments on 'conn' matrix                                         STATUS
% -------------------------------------------------------------------------
% - swap column 3 and 4 of conn ( quadrilateral element);                OK
% - no need for swaping when triangular element;                         OK
% - add 1 to all nodes of 'conn'. In comsol it starts on 0, in matlab,1; OK
% - attention to units! Use SI.                                          OK
% - check if the mesh has at least 6 nodes per wavelength          ON-GOING


% load manually imported comsol mesh
load('solid_rect_MESH.mat');

% plot mesh
figure(1)
femplot(l.x{2,2},l.elem{2,1},'Color' , 'black', 'Nodes', 'off', ...
'NodeNumbers' , 'on', 'ElementNumbers', 'on');

% mesh parameters
m = size(l.elem{2,1}, 1); % number of elements in the mesh
n = size(l.x{2,2}, 1);    % number of nodes in the mesh 

%% INTEGRATION POINTS WITHIN MASTER ELEMENT

% coordinates of integration points 
% (in the coordinate system of the master element)
% and the weight factors
xi = [ -1/sqrt(3)*e1 - 1/sqrt(3)*e2
        1/sqrt(3)*e1 - 1/sqrt(3)*e2
        1/sqrt(3)*e1 + 1/sqrt(3)*e2
       -1/sqrt(3)*e1 + 1/sqrt(3)*e2 ];
w  = [ 1
       1
       1
       1 ];

%% ASSEMBLE SYSTEM MATRICES IN ELEMENT LOOP 
% following the notation in my notes
% A - [1/Pa]      equivalent "mass matrix", term that multiplies d^2/dt^2 p 
% B - [1/(kg/m3)] equivalent "stiffness matrix", term that multiplies  p 
% d - [m/s2]      RHS term "external forces"

% neglect the intertial terms so d^2/dt^2 p = 0
% A = zeros(n, n); % zero matrix of nxn scalars 

B = zeros(n, n); % zero matrix of nxn scalars

% copy to shorter variable names
conn = l.elem{2,1};
x    = l.x{2,2};

   for e = 1:m % loop over all elements

    iie = conn(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop                    
    Be = zeros(4, 4);        % zero matrix of 4x4 scalars

    for k = 1:4 % loop over integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ 1/4*(1-xi1)*(1-xi2)
               1/4*(1+xi1)*(1-xi2)
               1/4*(1+xi1)*(1+xi2)
               1/4*(1-xi1)*(1+xi2) ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ -1/4*(1-xi2)*e1 - 1/4*(1-xi1)*e2
                      1/4*(1-xi2)*e1 - 1/4*(1+xi1)*e2
                      1/4*(1+xi2)*e1 + 1/4*(1+xi1)*e2
                     -1/4*(1+xi2)*e1 + 1/4*(1-xi1)*e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Be = Be + w(k) *(1/rho)*  dot(gradNe,gradNe') * thick * det(J);     
   
    end % end of interation point loop
   
% assembly
    B(iie, iie) = B(iie, iie) + Be;
end % end of element loop

%% APPLY BOUNDARY CONDITIONS BY PARTITIONING

% Manually select nodes by looking at the femplot
bottom = [3 6 10 14]; % nodes on the bottom boundary of the mesh
top    = [11 15 19 22]; % nodes on the top boundary of the mesh
left   = [1 4 7]; % nodes on the left boundary of the mesh
right  = [18 23 24]; % nodes on the right boundary of the mesh

  
% Prescribed nodes
iipresc = [21 2]; % conn [6 8] as in serendipity fashion [1 2 3 4 5 6 7 8]
% nodes where natural (Neumann) b.c. are applied ( free nodes:v=0).
% [bottom top]

% % the remaining nodes ( free nodes + Neumann BC v=0)
% fp = setdiff(1:n, pp);
% 
% % stiffness matrix
% Bpp = B(pp, pp);
% Bfp = B(fp, pp);
% Bpf = B(pp, fp);  
% Bff = B(fp, fp);

%% ELIMINATING DEPENDENT NODES

% constrain relation of periodicity 
% p(right)  = p(left)   +  (p(presc(1)) - p(presc(2))) ones()
% or
% p(right)  = p(left)   +  (p(6)        - p(8)       ) ones()

    % dependent nodes of periodic bc: u(iid)= u(iii)
    iid = [right];   % excluding corners
    
    % independent nodes of periodic bc
    iii = [left]; % excluding corners
    

    % safety check
    if(length(iii)~= length(iid)) 
        error("The periodic bc must connect same number of nodes");
    end

% unconstrained nodes ( nodes not at the boundary)
iiu = setdiff(1:n, [iipresc iii iid]); % exclude corners                           
iiu = [iipresc iiu];                   % include corners as first elements


% reordering mass and stiffness elements as [iiu iii iid]
    % stiffness tensor matrix
    % K_reordered =  Kuu Kui Kud   
    %                Kiu Kii Kid  
    %                Kdu Kdi Kdd  
    Buu = B(iiu, iiu); Bui = B(iiu, iii); Bud = B(iiu, iid);  
    Biu = B(iii, iiu); Bii = B(iii, iii); Bid = B(iii, iid);  
    Bdu = B(iid, iiu); Bdi = B(iid, iii); Bdd = B(iid, iid);  

% exclude prescibed nodes again to stick with notes notation. Note the
% prescibed nodes, 6 and 8 (based on serendipity conn fashion), are treated
% separetly for the projection matrix.
iiu([1 2])=[]; 

% projection matrix components
O_1u = zeros(1,length(iiu));
O_u1 = O_1u';

O_1L = zeros(1,length(left));
O_L1 = O_1L';

O_uL = zeros(length(iiu),length(left));
O_Lu = O_uL';

I_uu = eye(length(iiu),length(iiu));
I_LL = eye(length(left),length(left));

I_L1 = ones(length(left),1);


% projection matrix
%     p_6    = T * p_6
%     p_8          p_8
%     p_un         p_un
%     p_in         p_in
%     p_de

T  = [  1       0      O_1u    O_1L
        0       1      O_1u    O_1L
       O_u1    O_u1    I_uu    O_uL
       O_L1    O_L1    O_Lu    I_LL
       I_L1   -I_L1    O_Lu    I_LL];

% reduced stiffness tensor matrix
B_ =  T'*[Buu Bui Bud;
          Biu Bii Bid;
          Bdu Bdi Bdd]*T;




% set of remaining nodes
iire = [iipresc iiu iii];

% %% PRESCRIBED AND FREE NODES SPLIT
% %|__NOTES_NOTATION_|____ DESCRIPTION_______________________________________
% %|                 |
% %|        pp       | prescribed nodes in the fluid phase
% %|_________________|_______________________________________________________
% %|                 |
% %|        pf       | free nodes in the fluid phase                                                      
% %|                 |
% %|_________________|_______________________________________________________
% %--------------------------------------------------------------------------
% 
% fluid
    %first elements are the prescribed nodes by construction
    pp  = get_ind(iire,iipresc); % = [1 2] 
    fp  = setdiff(1:length(iire),pp);



%% PARTITIONING

% solid
    % stiffness matrix
    % B =  Bpp Bpf   
    %      Bfp Bff 
    B_pp_pp   = B_(pp , pp ); B_pp_fp   = B_(pp , fp );
    B_fp_pp   = B_(fp , pp ); B_fp_fp   = B_(fp , fp );
    
    
%% VALUES OF THE PRESCRIBED DISPLACEMENTS
% prescribe pressure to zero (open)
% pp  = [ PL*ones(length(left),1) ; PR*ones(length(right),1)]; 


% reference position vector
xR = (   x(1) + x(24)   )/2;
% xR = (   x(1) + x(1)   )/2;

% ones
ones_p = ones(length(pp),1);

% precribed fluid points are [left_f right_f]
    % MODEL INPUT
    pM     = -100;
    gradpM =  100*(e1+ 2*e2);
    % first-order comp. homogenization
    p_pp   =  pM *ones(length(iipresc),1) + ...
              dot(gradpM, x(iipresc)-xR*ones_p);
% Imposed external acceleration dv/dt = zero inward the fluid domain along
% [bottom top] nodes, i.e., hard wall.
% Imposed external acceleration = zero along inner (free) nodes.
% Hence, the sum of [bottom top] nodes plus [inner] nodes results in
% [bottom top inner] which is equal to iff=[allnodes]-[left right]. 'Minus  
% sign' here should be interpreted in a set sense (set difference).
% RHS is therefore a vector of zeros on piif nodes.
d_fp  = zeros(length(fp), 1);


%% SOLVE STATIONARY (PRESSURE) PROBLEM

% declare the free pressure nodes
% pf  = zeros(length(iif),1) ;

% solve right problem
p_pf = -inv(B_fp_fp)*B_fp_pp*p_pp;


%% ASSIGN SOLUTION TO REDUCED PRESSURE VECTOR SOLUTION

% declare
p_red     = zeros(length(pp)+length(fp),1);

% prescribed nodes
p_red(pp) = p_pp;

% free nodes, modal pressure shape
p_red(fp) = p_pf;

%% ASSIGN SOLUTION TO PRESSURE VECTOR SOLUTION

p_aux = T*p_red;
p([iipresc iiu iii iid])  = p_aux;
p=p.';
%% PLOT MODE SHAPE (NON NORMALIZED)
figure(2)
clf
% plot the pressure solution
scatter(dot(x,e1),dot(x,e2),1000,p,'filled');
% colorbar('Ticks',[-1,-.5,0,.5 1]);
colorbar;
% caxis([-1 1])
colormap('jet');
axis equal
hold on
femplot(x, conn, 'Color', 'r');
title("stationary pressure field" );
hold off

%% PLOT LINE CUTLINE
figure(3)
cutlinenodes = find(abs((dot(x,e2)))<0.005);
clf
scatter(dot(x(cutlinenodes),e1),p(cutlinenodes),'filled')
% yticks([-140:20:60])
grid on;
