%% DESCRIPTION
%--------------------------------------------------------------------------
% 04/04/2022 stationary study of fluid-solid rectangle.
% This case is relevant because it handles the macroscopic pressure
% gradient in y direction. If I were to use periodicity of pressure, then
% the gradxp is zero.
% R. Liupekevicius Carnielli
%
% based com "compression_solution.m" from ACCM
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%      (u4).__________(periodic )____________. (u3)
%          |                                |
%          |           solid(tag=3)         |
%(periodic)|                                | (periodic bc)
%          |________ (f-s interface)________|
%          |                                |
%          |                                |
%  (pL)    |           fluid(tag=2)         |  (pR: uniform pressure. homog. bc))
%          |                                |
%          |                                |
%          |________ (f-s interface)________|
%          |                                |
%(periodic)|           solid(tag=1)         | (periodic bc)
%          |                                |
%     (u1).|__________(periodic)____________|. (u2)
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
% What are the low frequency modes ?
%
% Make sure you also have same element type in comsol.
%--------------------------------------------------------------------------
clear; close all;
tic
%% ADD TENSORLAB & SHARED FUNCTIONS & MESHES TO PATH
path(path,['C:\GitLab\Tools\TensorLab20\TensorLab20']);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\shared_functions"]);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\meshes"]);


%% PARAMETERS DEFINITION
disp('codeblock: PARAMETERS DEFINITION')

% geometric:
% H = 1; % height, in [m]
% L = 2; % length, in [m]
thick = 1;  % thickness (out-of-plane direction), in [m]


%--------------------------------------------------------------------------
% solid material properties
% G        = 1000e6; % in [Pa]
% kappa    = 3000e6; % in [Pa]
% rhos     = 8000;   % in [kg/m3]
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% solid material properties PU from Mirka ch3 (IN TERMS OF E AND nu)
rhos     = 1000;   % in [kg/m3]
E        = 1e6;    % [Pa]
nu       = 0.4;    % []    
G        = E/2/(1+  nu);
kappa    = E/3/(1-2*nu);
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% fluid properties
rhof     = 1.225; % [kg/m3]
c        = 343 ;  % [m/s]
% pA       = 101325; [Pa]   % not used for order 1 equations.
%--------------------------------------------------------------------------


% define the basis
b  = {'e1'; 'e2'};
ee = cartesianbasis2d(b{1}, b{2});
e1 = ee(1);
e2 = ee(2);


% determine material coefficient tensor
I   = identity(2, ee);
I4  = identity(4, ee);
I4S = 1/2 * (I4 + rtranspose(I4));
C   = kappa *I*I + 2*G * (I4S - 1/3*I*I);
toc
tic


             
%% MANUALLY IMPORT COMSOL MESH AS STRUCT WITH CELL ELEMENTS
disp('codeblock: MESH')
%--------------------------------------------------------------------------
% open txt file comsol_mesh_solid_rect.mphtxt and copy/paste the node 
% position and connectivity matrix.
% -------------------------------------------------------------------------
%  Comments on 'conn' and 'tag' matrices for elements and edges      STATUS
% -------------------------------------------------------------------------
% - swap column 3 and 4 of conn ( quadrilateral element);                OK
% - no need for swaping when triangular element;                         OK
% - add 1 to all nodes of 'conn' (maybe 'tag' too). In comsol it starts
% in 0, whereas in matlab, it starts in 1;                               OK
% - attention to units! Use SI.                                          OK
% - check if the mesh has at least 6 nodes per wavelength          ON-GOING
%
%
%
% -------------------------------------------------------------------------
%  Comments on edge/element tags in comsol mesh file                                
% -------------------------------------------------------------------------
% - list of fluid-solid edge tag: 4,6
% This means that if an edge has tag 4 or 6, than the edge is at the fluid-
% solid interface.
% - list of element type (fluid or solid) after 'conn' matrix
%--------------------------------------------------------------------------

% load manually imported comsol mesh (use same mesh as in solid case)
load('es_fluidsolid_rect_MESH.mat');
% transform to quadratic element
q=mesh2d_lin2qua(l);

% copy to local variables (please double click on 'mesh' in workspace to
% see the meaning of each cell)
x    = q.x{2,2};    % tensor
conn = q.elem{2,1}; % conn matrix of elements




% mesh parameters
m  = size(conn, 1); % number of elements in the mesh
n  = size(x, 1);    % number of nodes in the mesh 


% SOLID
    % solid elements
    % solid_elems = [find(mesh.elems('tag')==1);find(mesh.elems('tag')==3)].';
    solid_elems = union(find(q.elem{2,2}==1),find(q.elem{2,2}==3)).';
    % position vector of the solid nodes
%     xs          = x(solid_elems);
    % conn solid
    conns       = conn(solid_elems,:);
    % solid nodes
    nodes_s     = unique(reshape(conns,1,[]));
    % number of solid elements in the mesh
    ms          = size(conns,1);
    % number of solid nodes
    ns          = length(nodes_s);


% FLUID
    % fluid elements
    fluid_elems = setdiff(1:m,solid_elems);
    % position vector of the fluid nodes
%     xf          = x(fluid_elems);
    % conn fluid
    connf       = conn(fluid_elems,:);
    % fluid nodes
    nodes_f     = unique(reshape(connf,1,[]));
    % number of fluid elements in the mesh
    mf          = size(connf,1);
    % number of fluid nodes
    nf          = length(nodes_f);

    
%% INTERFACE NODES (MANUALLY CHECKING TAGS)

% FORMER APPROACH
% import edge tags from comsol MANUALLY
tags_itr_edges=[4 6];
%(make 'for' loop below if list is too long), here only 2 interfaces

% 'interface edges' are the indices of 'conn' matrix
itr_edges_tag4 = find(q.edge{2,2}==4).';
itr_edges_tag6 = find(q.edge{2,2}==6).';
itr_edges  = [itr_edges_tag4 itr_edges_tag6];
% these are edges(two nodes per edge) that form the interface.
% The edge number ( or edge 'element') is the index of 'conn'

% mesh.edge('conn')( interface_edges , :)
conn_edges = q.edge{2,1}; %temporary assignment
conni_tag4 = conn_edges(itr_edges_tag4,:);
conni_tag6 = conn_edges(itr_edges_tag6,:);
conni      = conn_edges(itr_edges,:);
clear conn_edges;

% Get interface nodes (2D)
% - reshape conn (edges) matrix on a concatenated vector
% - keep 'unique' nodes of concatenated vector, i.e. remove duplicates
% - interface_nodes is therefore the index of elements' conn matrix
nodes_itr4     = unique(reshape( conni_tag4,1,[] ) ).';
nodes_itr6     = unique(reshape( conni_tag6,1,[] ) ).';
nodes_itr      = unique(reshape( conni     ,1,[] ) ).';


% ALTERNATIVE ( however need of 'conni' matrix)
% nodes_itr =  intersect(nodes_s,nodes_f);

toc
tic


%__________________________________________________________________________
% INDEX TABLE
% FLUID (F)
% SOLID (S)
%
%        CURRENT_INDEX      
%               |  
%               V
%__________________________________Reordered____________________Reordered__
% CONVERTION | mesh   | nodes_s |   -----      |    dofs_re   |   ---
%    RULE    | mesh   | nodes_f |   nodes_f    |    nodes_f   |  nodes_f
%____________|__(FS) _|_(S)_(F)_|__(S)_____(F)_|___(S)___(F)__|__(S)_(F)__
%            |    1   |   1 1   |  dofs_c   1  |  dofs_pc  1  |    p  pp
%            |    2   |   2 2   |  dofs_un  2  |  dofs_un  2  |    f  ff
%   ORDER    |    .   |   . .   |  dofs_in  .  |  dofs_in  .  |
%            |    .   |   . .   |  dofs_de  .  |           .  |
%            |    .   |   . .   |           nf |           nf | 
%            |    n   |  ns nf  |              |              |
%____________|___ ____|_________|______________|______________|____________


%%  4 INTEGRATION POINTS WITHIN MASTER ELEMENT
disp('codeblock: ASSEMBLING')
%--------------------------------------------------------------------------
% Notation from TensorLab
% quadratic quadrilateral element:
%         4----7----3
%         |         |
%         8         6
%         |         |
%         1----5----2
%      then CONN = [1 2 3 4 5 6 7 8].
%
% The coordinates of integration points x (in the coordinate system of the
% master element). Exact integration up to polynomial of order 5.
%          --------- 
%         | x  x  x |
%         | x  x  x |
%         | x  x  x |
%          ---------
% and the weight factors
%--------------------------------------------------------------------------

xi = [ -sqrt(3/5) *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 + sqrt(3/5) *e2
       -sqrt(3/5) *e1 + sqrt(3/5) *e2
                0 *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 +         0 *e2       
                0 *e1 + sqrt(3/5) *e2                
       -sqrt(3/5) *e1 +         0 *e2
                0 *e1 +         0 *e2   ];
w  = [ 25/81
       25/81
       25/81
       25/81
       40/81
       40/81
       40/81
       40/81
       64/81     ];

%% ASSEMBLE SYSTEM MATRICES IN A SOLID ELEMENT LOOP 
%--------------------------------------------------------------------------
% Knn and Mnn are only temporary matrix and they are not the matrices K and
% M from discretized form in the notes. Note their size is the total number
% of nodes which include the fluid nodes. It's only relevant property is
% that it preserves the node indexing.
%--------------------------------------------------------------------------

Knn = zeros(2, ee, n, n); % zero matrix of nxn 2nd order tensors in basis ee

   for e = 1:size(conns,1) % loop over all solid elements

    iie = conns(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop
    Ke = zeros(2, ee, 8, 8); % zero matrix of 8x8 2nd order tensors
                             % in basis ee.
 
    for k = 1:length(w) % loop over 9 integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ -1/4*(1-xi1  )*(1-xi2  )*(1+xi1+xi2)
               -1/4*(1+xi1  )*(1-xi2  )*(1-xi1+xi2)
               -1/4*(1+xi1  )*(1+xi2  )*(1-xi1-xi2)
               -1/4*(1-xi1  )*(1+xi2  )*(1+xi1-xi2)
                1/2*(1-xi1^2)*(1-xi2  )
                1/2*(1+xi1  )*(1-xi2^2)
                1/2*(1-xi1^2)*(1+xi2  )
                1/2*(1-xi1  )*(1-xi2^2)           ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ - 1/4 * (-1 + xi2   )* (2*xi1 +   xi2) *e1 ...
                     - 1/4 * (-1 + xi1   )* (xi1   + 2*xi2) *e2
                     - 1/4 * (2*xi1 - xi2)* (-1    + xi2  ) *e1 ...
                     - 1/4 * (1 + xi1    )* (xi1   - 2*xi2) *e2
                       1/4 * (1 + xi2    )* (2*xi1 + xi2  ) *e1 ...
                     + 1/4 * (1 + xi1    )* (xi1   + 2*xi2) *e2
                       1/4 * (2*xi1 - xi2)* (1     + xi2  ) *e1 ...
                     + 1/4 * (-1 + xi1   )* (xi1 - 2*xi2  ) *e2
                                      xi1 * (-1 + xi2     ) *e1 ...
                                    + 1/2 * (-1 + xi1^2   ) *e2
                                      1/2 * ( 1 - xi2^2   ) *e1 ...
                                    - xi2 * ( 1 + xi1     ) *e2
                                    - xi1 * ( 1 + xi2     ) *e1 ...
                                    + 1/2 * ( 1 - xi1^2   ) *e2
                                      1/2 * (-1 + xi2^2   ) *e1 ...
                                    + xi2 * (-1 + xi1     ) *e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Ke = Ke + w(k) *  dot(gradNe, C, gradNe') * thick * det(J);
         
   
    end % end of interation point loop
   
   % assembly
    Knn(iie, iie) = Knn(iie, iie) + Ke;
   end % end of element loop

% assignment actual mass and stiffness matrices
K=Knn(nodes_s,nodes_s);
clear Knn;

%% ASSEMBLE SYSTEM MATRICES IN A FLUID ELEMENT LOOP 
%--------------------------------------------------------------------------
% following the notation in my notes
% A - [1/Pa]      equivalent "mass matrix", term that multiplies d^2/dt^2 p 
% B - [1/(kg/m3)] equivalent "stiffness matrix", term that multiplies  p 
% d - [m/s2]      RHS term "external forces"
%--------------------------------------------------------------------------

Bnn = zeros(n, n); % zero matrix of nxn scalars


   for e = 1:size(connf,1)% loop over all fluid elements

    iie = connf(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop       
    Be = zeros(8, 8);        % zero matrix of 8x8 scalars

    for k = 1:length(w) % loop over 9 integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ -1/4*(1-xi1  )*(1-xi2  )*(1+xi1+xi2)
               -1/4*(1+xi1  )*(1-xi2  )*(1-xi1+xi2)
               -1/4*(1+xi1  )*(1+xi2  )*(1-xi1-xi2)
               -1/4*(1-xi1  )*(1+xi2  )*(1+xi1-xi2)
                1/2*(1-xi1^2)*(1-xi2  )
                1/2*(1+xi1  )*(1-xi2^2)
                1/2*(1-xi1^2)*(1+xi2  )
                1/2*(1-xi1  )*(1-xi2^2)           ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ - 1/4 * (-1 + xi2   )* (2*xi1 +   xi2) *e1 ...
                     - 1/4 * (-1 + xi1   )* (xi1   + 2*xi2) *e2
                     - 1/4 * (2*xi1 - xi2)* (-1    + xi2  ) *e1 ...
                     - 1/4 * (1 + xi1    )* (xi1   - 2*xi2) *e2
                       1/4 * (1 + xi2    )* (2*xi1 + xi2  ) *e1 ...
                     + 1/4 * (1 + xi1    )* (xi1   + 2*xi2) *e2
                       1/4 * (2*xi1 - xi2)* (1     + xi2  ) *e1 ...
                     + 1/4 * (-1 + xi1   )* (xi1 - 2*xi2  ) *e2
                                      xi1 * (-1 + xi2     ) *e1 ...
                                    + 1/2 * (-1 + xi1^2   ) *e2
                                      1/2 * ( 1 - xi2^2   ) *e1 ...
                                    - xi2 * ( 1 + xi1     ) *e2
                                    - xi1 * ( 1 + xi2     ) *e1 ...
                                    + 1/2 * ( 1 - xi1^2   ) *e2
                                      1/2 * (-1 + xi2^2   ) *e1 ...
                                    + xi2 * (-1 + xi1     ) *e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Be = Be + w(k) *(1/rhof)*  dot(gradNe,gradNe') * thick * det(J);     
   
    end % end of interation point loop
   
% assembly
    Bnn(iie, iie) = Bnn(iie, iie) + Be;
   end % end of fluid element loop

% define actual stiffness and mass matrices.
B=Bnn(nodes_f,nodes_f);
clear Bnn;


%% INTEGRATION POINTS WITHIN MASTER EDGE (ELEMENT)
%--------------------------------------------------------------------------
% use here the same shape functions for both fluid and solid edges
% the edges have three nodes refered as 1, 2 and 3 below.
%
%              solid              
%      1--------3---------2  (master nodes)
%       --x-----x-----x--    (integration points)

%              fluid             | (solid normal)   
%                                V 
% then CONN = [1 3 2].
% coordinates of integration points sketched as x.  Exact integration up
% to polynomial of order 5.
%
% in the coordinate system of the master element and the weight factors
%--------------------------------------------------------------------------

xi = [ -sqrt(3/5)
              0
        sqrt(3/5) ];
w  = [ 5/9
       8/9
       5/9        ];

%% ASSEMBLE COUPLING MATRIX, LOOP ON INTERFACE EDGES 
%--------------------------------------------------------------------------
% use here the same shape functions for both fluid and solid elements
%--------------------------------------------------------------------------

Cnn = zeros(1, ee, n, n);        % zero matrix of nxn vectors


   for e = 1:size(conni,1) % loop over interface edges
    
    % edge e and its coordinates
    iie      = conni(e, :); % nodes of the current edge e
    xe       = x(iie);      % coordinates of these nodes
    
    % solid normal vector
    ve       = xe(3)-xe(1); % edge vector (of edge e)
    normal_s = ( dot(ve,e2)*e1 - dot(ve,e1)*e2)/...
                sqrt( dot(ve,e1)^2+dot(ve,e2)^2 ); % solid normal
%   normal_f = (-dot(ve,e2)*e1 + dot(ve,e1)*e2)/norm(ve); % fluid normal

    % compute element matrices in integration point loop
    Ce =  zeros(1, ee, 3, 3);        % 3x3 zero matrix of vectors

    for k = 1:length(w) % loop over 3 integration points 
        
       % xi(k) is the coordinate of the current integr. point
       % column of the shape functions
        Ne = [ -1/2* (1-xi(k))*    xi(k)
                     (1-xi(k))* (1+xi(k))
                1/2*    xi(k) * (1+xi(k))  ];

       % elementary length ds=|dx|d(xi)       =  1/2*norm(ve)*d(xi) 
       % elementary area   dA=|dx|d(xi) thick =  1/2*norm(ve)*thick*d(xi)
       
    
       % element matrix 
        Ce = Ce + w(k) * (Ne * Ne') * normal_s * 1/2*norm(ve)* thick ;
        
         
   
    end % end of interation point loop
   
% assembly
    Cnn(iie, iie) = Cnn(iie, iie) + Ce;

   end % end of interface edge loop

C = Cnn(nodes_s,nodes_f);
clear Cnn;

% % check nonzero elements seing the following matrices
% % "primitive conversion" to scalar for a matter of checking nonzero compon.
% Cnncheck = dot(Cnn,e1).^2 + dot(Cnn,e2).^2; 
% Cnncheck(nodes_itr,nodes_itr);
% Cnncheck(setdiff(1:n,nodes_itr),setdiff(1:n,nodes_itr));

toc
tic


%% NODE SORTING
disp('codeblock: NODE SORTING/ PBC PROJECTION/ PARTITIONING')

% Split the set of nodes 
% TODO: play with gmsh and automate the bc node sorting

% MANUALLY select nodes by looking at the femplot
% fluid
    % nodes on the left boundary of the mesh
    left_f    = [13 19 26 35 42 54 66 78 92]; 
    % nodes on the right boundary of the mesh
    right_f   = [386 387 389 391 393 396 398 399 402]; 
% solid
    % nodes on the bottom boundary of the mesh (excluding corners)
    bottom_s  = [3 5 10 15 22 29 39 49 61 73 87 104 119 136 156 176 ...
                 194 209 225 240 254 268 285 297 313 324 340 352 367];
    % nodes on the top boundary of the mesh (excluding corners)
    top_s     = [158 162 165 169 172 181 188 193 203 211 222 230 ...
                 237 249 258 267 279 291 302 312 323 336 347 359 370 ...
                 381 392 400 404]; 
    % nodes on the left boundary of the mesh (excluding corners)
    left_s    = [2 4 9 13 92 107 122 139]; 
    % nodes on the right boundary of the mesh (excluding corners)
    right_s   = [380 383 384 386 402 403 405 406];
    % corner nodes
    corners_s = [1 379 407 155];
    % prescribed corners
    pcorners = [corners_s(1) corners_s(2) corners_s(4)];


%% INDEX OF DISPLACEMENT/PRESSURE DOF IN NODES_S AND NODES_F
%--------------------------------------------------------------------------
% The matrices Mnn, Knn, Ann, Bnn, and Cnn have indices that correspond to
% the node number, i.g., line 1 are the terms related to node 1.
% The matrices M, K, A, B, and C, however, have indices that do not 
% correspond to the node number. Their indices correspond to the
% displacements degree of freedom or pressure degrees of freedom.
% Suppose, for example, you want to find the index in matrix M that
% correspond to the bottom-left corner displacement dof. This corner node
% is stored in the vector [corner], say first entry, corner(1). The index
% we are looking for, sometimes refered as 'new' index, is found using the
% following command:
% 
% find(nodes_s==corner(1)).
%
% get_ind() was implemented for this operation
%--------------------------------------------------------------------------

% fluid
    % left nodes
    doff_l = get_ind(nodes_f, left_f   ); 
    % right nodes
    doff_r = get_ind(nodes_f, right_f  ); 
    % interface nodes
    doff_i = get_ind(nodes_f, nodes_itr);
% solid 
    % bottom nodes
    dofs_b = get_ind(nodes_s, bottom_s ); 
    % top nodes
    dofs_t = get_ind(nodes_s, top_s    ); 
    % left nodes
    dofs_l = get_ind(nodes_s, left_s   ); 
    % right nodes
    dofs_r = get_ind(nodes_s, right_s  ); 
    % interface nodes
    dofs_i = get_ind(nodes_s, nodes_itr);
    % corner nodes
    dofs_c = get_ind(nodes_s, corners_s);
    % prescribed corners
    dofs_pc = get_ind(nodes_s, pcorners);

 
%--------------------------------------------------------------------------
% TODO:afterwards, check which sets (above) of dof were actually used.
%--------------------------------------------------------------------------

%__________________________________________________________________________
% INDEX TABLE
% FLUID (F)
% SOLID (S)
%
%                   CURRENT_INDEX      
%                          |  
%                          V
%__________________________________Reordered____________________Reordered__
% CONVERTION | mesh   | nodes_s |   -----      |    dofs_re   |   ---
%    RULE    | mesh   | nodes_f |   nodes_f    |    nodes_f   |  nodes_f
%____________|__(FS) _|_(S)_(F)_|__(S)_____(F)_|___(S)___(F)__|__(S)_(F)__
%            |    1   |   1 1   |  dofs_c   1  |  dofs_pc  1  |    p  pp
%            |    2   |   2 2   |  dofs_un  2  |  dofs_un  2  |    f  ff
%   ORDER    |    .   |   . .   |  dofs_in  .  |  dofs_in  .  |
%            |    .   |   . .   |  dofs_de  .  |           .  |
%            |    .   |   . .   |           nf |           nf | 
%            |    n   |  ns nf  |              |              |
%____________|___ ____|_________|______________|______________|____________


%% ELIMINATING DEPENDENT NODES OF SOLID PHASE
%--------------------------------------------------------------------------
% constrain relation of periodicity 
% u(top)    = u(bottom) +  (u(coners(4))-u(coners(1))) ones()
% u(right)  = u(left)   +  (u(coners(2))-u(coners(1))) ones()
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
    % dependent dofs of periodic bc
    dofs_de = [dofs_t dofs_r];   % [top right], excluding corners
    
    % independent nodes of periodic bc
    dofs_in = [dofs_b dofs_l];    %[bottom left], excluding corners
    

    % safety check
    if(length(dofs_in)~= length(dofs_de)) 
        error("The periodic bc must connect same number of nodes");
    end

% unconstrained nodes ( nodes not at the boundary)
dofs_un = setdiff(1:ns, [dofs_in dofs_de dofs_c]); % exclude corners                           
% include corners as first elements to follow the notes notation
dofs_un = [dofs_c dofs_un];           
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% reordering mass, stiffness, and coupling matrix components


% stiffness tensor matrix
% K=  Kuu Kui Kud   
%     Kiu Kii Kid  
%     Kdu Kdi Kdd  
Kuu=K(dofs_un, dofs_un); Kui=K(dofs_un, dofs_in); Kud=K(dofs_un, dofs_de);  
Kiu=K(dofs_in, dofs_un); Kii=K(dofs_in, dofs_in); Kid=K(dofs_in, dofs_de);  
Kdu=K(dofs_de, dofs_un); Kdi=K(dofs_de, dofs_in); Kdd=K(dofs_de, dofs_de);  
% coupling tensor matrix
% C=  Cu    
%     Ci   
%     Cd  
Cu=C(dofs_un,:);
Ci=C(dofs_in,:);
Cd=C(dofs_de,:);
%--------------------------------------------------------------------------

%__________________________________________________________________________
% INDEX TABLE
% FLUID (F)
% SOLID (S)
%
%                               CURRENT_INDEX      
%                                      |  
%                                      V
%__________________________________Reordered____________________Reordered__
% CONVERTION | mesh   | nodes_s |   -----      |    dofs_re   |   ---
%    RULE    | mesh   | nodes_f |   nodes_f    |    nodes_f   |  nodes_f
%____________|__(FS) _|_(S)_(F)_|__(S)_____(F)_|___(S)___(F)__|__(S)_(F)__
%            |    1   |   1 1   |  dofs_c   1  |  dofs_pc  1  |    p  pp
%            |    2   |   2 2   |  dofs_un  2  |  dofs_un  2  |    f  ff
%   ORDER    |    .   |   . .   |  dofs_in  .  |  dofs_in  .  |
%            |    .   |   . .   |  dofs_de  .  |           .  |
%            |    .   |   . .   |           nf |           nf | 
%            |    n   |  ns nf  |              |              |
%____________|___ ____|_________|______________|______________|____________


%--------------------------------------------------------------------------
% exclude corners again to stick with notes notation, they will be treated
% separately by the projection matrix
dofs_un([1 2 3 4])=[]; 
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% projection matrix submatrices 
O_1u = zeros(1,length(dofs_un));
O_u1 = O_1u';

O_1B = zeros(1,length(dofs_b));
O_B1 = O_1B';

O_1L = zeros(1,length(dofs_l));
O_L1 = O_1L';

O_uB = zeros(length(dofs_un),length(dofs_b));
O_Bu = O_uB';

O_uL = zeros(length(dofs_un),length(dofs_l));
O_Lu = O_uL';

O_BL = zeros(length(dofs_b),length(dofs_l));
O_LB = O_BL';

I_uu = eye(length(dofs_un),length(dofs_un));
I_BB = eye(length(dofs_b),length(dofs_b));
I_LL = eye(length(dofs_l),length(dofs_l));

I_B1 = ones(length(dofs_b),1);
I_L1 = ones(length(dofs_l),1);

% projection matrix
%     u_1   = T * u_1
%     u_2         u_2
%     u_3         u_4
%     u_4         u_un
%     u_un        u_in
%     u_in
%     u_de

% projection matrix
T  = [  1       0      0      O_1u    O_1B    O_1L
        0       1      0      O_1u    O_1B    O_1L
       -1       1      1      O_1u    O_1B    O_1L
        0       0      1      O_1u    O_1B    O_1L
       O_u1    O_u1   O_u1    I_uu    O_uB    O_uL
       O_B1    O_B1   O_B1    O_Bu    I_BB    O_BL
       O_L1    O_L1   O_L1    O_Lu    O_LB    I_LL
      -I_B1    O_B1   I_B1    O_Bu    I_BB    O_BL
      -I_L1    I_L1   O_L1    O_Lu    O_LB    I_LL];
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% reduced stiffness tensor matrix (dependent nodes eliminated)
K_ =  T'*[Kuu Kui Kud;
          Kiu Kii Kid;
          Kdu Kdi Kdd]*T;

% reduced mass tensor matrix (dependent nodes eliminated)
C_ =  T'*[Cu;
          Ci;
          Cd];
%--------------------------------------------------------------------------

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% question: projection with T' or T.' ? TODO: test both
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%--------------------------------------------------------------------------
% set of remaining nodes, this is the vector that contains the rules for
% changing the indices from 'before projection' to 'after projection'.
         
dofs_re = [dofs_c(1)  dofs_c(2)  dofs_c(4)  dofs_un dofs_in];
ns_re   = length(dofs_re);
%--------------------------------------------------------------------------

%__________________________________________________________________________
% INDEX TABLE
% FLUID (F)
% SOLID (S)
%
%                                               CURRENT_INDEX      
%                                                      |  
%                                                      V
%__________________________________Reordered____________________Reordered__
% CONVERTION | mesh   | nodes_s |   -----      |    dofs_re   |   ---
%    RULE    | mesh   | nodes_f |   nodes_f    |    nodes_f   |  nodes_f
%____________|__(FS) _|_(S)_(F)_|__(S)_____(F)_|___(S)___(F)__|__(S)_(F)__
%            |    1   |   1 1   |  dofs_c   1  |  dofs_pc  1  |    p  pp
%            |    2   |   2 2   |  dofs_un  2  |  dofs_un  2  |    f  ff
%   ORDER    |    .   |   . .   |  dofs_in  .  |  dofs_in  .  |
%            |    .   |   . .   |  dofs_de  .  |           .  |
%            |    .   |   . .   |           nf |           nf | 
%            |    n   |  ns nf  |              |              |
%____________|___ ____|_________|______________|______________|____________


%% PRESCRIBED AND FREE NODES SPLIT
%--------------------------------------------------------------------------
% INDEX OF DISPLACEMENT DOF IN NODES_S (ELIMINATED DEPENDENT DOFS)
% same index change procedure must be executed but only in the solid phase.

%|__NOTES_NOTATION_|____ DESCRIPTION_______________________________________
%|                 |
%|        p        | prescribed nodes in the solid phase
%|_________________|_______________________________________________________
%|                 |
%|        f        | free nodes in the solid phase                                                      
%|                 |
%|_________________|_______________________________________________________
%|                 |
%|        p'       | prescribed nodes in the fluid phase
%|_________________|_______________________________________________________
%|                 |
%|        f'       | free nodes in the fluid phase
%|_________________|_______________________________________________________
%--------------------------------------------------------------------------

% solid
p  = get_ind(dofs_re,dofs_pc); % = [1 2 3] by construction
f  = setdiff(1:ns_re,p);

% fluid
pp = [doff_l doff_r];
fp = setdiff(1:nf,pp);

%__________________________________________________________________________
% INDEX TABLE
% FLUID (F)
% SOLID (S)
%
%                                                            CURRENT_INDEX      
%                                                                   |  
%                                                                   V
%__________________________________Reordered____________________Reordered__
% CONVERTION | mesh   | nodes_s |   -----      |    dofs_re   |   ---
%    RULE    | mesh   | nodes_f |   nodes_f    |    nodes_f   |  nodes_f
%____________|__(FS) _|_(S)_(F)_|__(S)_____(F)_|___(S)___(F)__|__(S)_(F)__
%            |    1   |   1 1   |  dofs_c   1  |  dofs_pc  1  |    p  pp
%            |    2   |   2 2   |  dofs_un  2  |  dofs_un  2  |    f  ff
%   ORDER    |    .   |   . .   |  dofs_in  .  |  dofs_in  .  |
%            |    .   |   . .   |  dofs_de  .  |           .  |
%            |    .   |   . .   |           nf |           nf | 
%            |    n   |  ns nf  |              |              |
%____________|___ ____|_________|______________|______________|____________

%% PARTITIONING

%--------------------------------------------------------------------------
% solid
    % stiffness tensor matrix
    % K =  Kpp Kpf   
    %      Kfp Kff 
    K_p_p   = K_(p , p ); K_p_f   = K_(p , f );
    K_f_p   = K_(f , p ); K_f_f   = K_(f , f );
    
% fluid 
    % 'stiffness' scalar matrix 
    % B =  Bp'p' Bp'f'   
    %      Bf'p' Bf'f' 
    B_pp_pp = B (pp, pp); B_pp_fp = B (pp, fp);
    B_fp_pp = B (fp, pp); B_fp_fp = B (fp, fp);

% coupling tensor matrix
    % C =  Cpp' Cpf'   
    %      Cfp' Cff' 
    C_p_pp  = C_(p , pp);C_p_fp  = C_(p , fp);
    C_f_pp  = C_(f , pp);C_f_fp  = C_(f , fp);
%--------------------------------------------------------------------------

toc
tic


%% TRANFORM TENSOR MATRIX TO EQUIVALENT SCALAR MATRIX
disp('codeblock: TRANFORM TENSOR MATRIX TO EQUIVALENT SCALAR MATRIX')

%--------------------------------------------------------------------------
% stiffness matrix Kff
sK_p_p = tm2sm_v( K_p_p, ee , 2 ); sK_p_f = tm2sm_v( K_p_f, ee , 2 );
sK_f_p = tm2sm_v( K_f_p, ee , 2 ); sK_f_f = tm2sm_v( K_f_f, ee , 2 );

% coupling matrix C
sC_p_pp = tm2sm_v(C_p_pp, ee, 1); sC_p_fp = tm2sm_v(C_p_fp, ee, 1);
sC_f_pp = tm2sm_v(C_f_pp, ee, 1); sC_f_fp = tm2sm_v(C_f_fp, ee, 1);

% zero matrices declaration directly as equivalent scalar matrix
%O_pp_p = sparse(zeros( length(pp) , size(ee,1)*length(p)));
O_fp_p = sparse(zeros( length(fp) , size(ee,1)*length(p)));
O_fp_f = sparse(zeros( length(fp) , size(ee,1)*length(f)));
O_pp_f = sparse(zeros( length(pp) , size(ee,1)*length(f)));
%--------------------------------------------------------------------------

toc
tic


%% ASSEMBLE NONSYMMETRIC SYSTEM OF EQUATIONS 
disp('codeblock: SOLVE')

%--------------------------------------------------------------------------
% cond of the following matrices( with the term rhof*c2*) is lower
% % mu matrix
% muFF  = [  sK_f_f              sC_f_fp ;
%             O_fp_f          rhof*c^2*B_fp_fp    ];  
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% mu submatrices
% muPP  = [  sK_p_p              sC_p_pp ;
%             O_pp_p              B_pp_pp    ];  

% mu matrix
muPF  = [  sK_p_f              sC_p_fp ;
            O_pp_f              B_pp_fp    ];  
% mu matrix
muFP  = [  sK_f_p              sC_f_pp ;
            O_fp_p              B_fp_pp    ];  
% mu matrix
muFF  = [  sK_f_f              sC_f_fp ;
            O_fp_f              B_fp_fp    ];  
%--------------------------------------------------------------------------




%% PRESCRIBED DISPLACEMENT, PRESSURE AND ITS GRADIENTS

%--------------------------------------------------------------------------
% reference position vector
xR = (   x(corners_s(1)) + x(corners_s(3))   )/2;

% ones
ones_u = ones(length(p) ,1);
ones_p = ones(length(pp),1);
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% prescribed solid points are [corners_s]
    % MODEL INPUT
    uM     = 0*(e1+e2);
    graduM = 0.0 *e1*e1 + 0.0   *e1*e2 +...
             0.0 *e2*e1 + 0.0 *e2*e2;
    % first-order comp. homogenization, prescribed disp. at the corners
    % note input index of x() is always "mesh" type.
    u_p    =  uM *ones_u + dot(graduM.', x(pcorners)-xR*ones_u);
    % transform to equivalent scalar (no need for sparsicity on a vector)
    su_p   = full(tm2sm_v(u_p, ee, 1));

if length(p)~=length(pcorners);         error('something wrong');end
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% precribed fluid points are [left_f right_f]
    % MODEL INPUT
    pM     = -1;
    gradpM =  100*(e1+2*e2);
    % first-order comp. homogenization
    p_pp   =  pM *ones(length(pp),1) + ...
              dot(gradpM, x([left_f right_f])-xR*ones_p);
if length(pp)~=length([left_f right_f]); error('something wrong');end

% stack both in a hybrid state vector containing disp. and pressure
wP = [  su_p
         p_pp];

clear su_p;


% %% SOLVE RIGHT/LEFT LINEAR SYSTEM OF EQUATIONS
%--------------------------------------------------------------------------
% right stationary hybrid state vector
wr_f = - inv(muFF  ) * muFP   * wP;

% left stationary hybrid state vector
wl_f = - inv(muFF.') * muPF.' * wP;
%--------------------------------------------------------------------------



% %% ASSIGN DISP. SOLUTION TO TENSOR VECTOR
%--------------------------------------------------------------------------
% 'ured' is  reduced displacement thanks
% to the elimitation of the dependend points due to periodic bc
%
%  length(ured) = length([dofs_pc; dofs_un; dofs_in])
%  length(u   ) = length([dofs_c ; dofs_un; dofs_in; dofs_de])
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% free (reduced) displacements is the right stationary solution
    % declare
    % nodes are between index 1 to 2*length(f)
    ured_f = wr_f(1:2:2*length(f)-1)*e1+ wr_f(2:2:2*length(f))*e2;
% total reduced displacements (indices ordered as [dofs_un dofs_in])
    % declare ured(ordered: [unconstrained independent] nodes)
    ured = zeros(1, ee, length(p)+length(f), 1);
    % assign prescribed nodes
    ured(p,:) = u_p;
    % assign free nodes
    ured(f,:) = ured_f;
    % free unused variables
    clear ured_f; clear u_p;
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% declare displacement tensor vectors (total # of dofs)
vu     = zeros(1, ee, ns, 1);
vu_aux = zeros(1, ee, ns, 1); 
% assign. Note vu_aux is in order [dofs_c dofs_un dofs_in dofs_de]
vu_aux = T*ured;
% assign to vu. Note vu is in [nodes_s] order
vu([dofs_c dofs_un dofs_in dofs_de])  = vu_aux;
% free unused variables
clear vu_aux; clear ured;

% prepare for plot: length(vu_plot)>length(vu)
    % declare disp. vector for plot -  whole mesh
    vu_plot = zeros(1, ee, n, 1);
    % assign vu to vector for plot, only solid nodes are non-zero
    vu_plot(nodes_s) = vu;
    % assign macro displacement to fluid nodes
%     vu_plot(nodes_f) = uM*ones(nf,1);
%--------------------------------------------------------------------------



% %% ASSIGN PRESSURE SHAPE MODES TO SCALAR VECTOR vp
%--------------------------------------------------------------------------
% extract pressure values from state vector wr_f
    % free pressures 
    p_fp = wr_f(end-length(fp)+1:end);

% declare pressure scalar vector
vp = zeros(nf,1);

% assign to scalar vector solution of size
vp(pp,:)=p_pp;
vp(fp,:)=p_fp;

% prepare for plot length(vp_plot)>length(vp)
    % declare pressure scalar vector for plot - whole mesh vector
    vp_plot = zeros(n,1);
    
    % assign vp to vector for plot
    vp_plot(nodes_f) = vp;
%--------------------------------------------------------------------------

toc
tic

disp('codeblock: PLOT')

% %% PLOT PRESSURE/DISP SHAPE (RESCALED DISP.)

%--------------------------------------------------------------------------
figure(4)
clf
% plot the pressure solution and displacement
scatter(dot(x,e1),dot(x,e2),150,vp_plot,'filled')
colorbar;
pmax = max(abs(vp_plot));
caxis([-pmax pmax]);
colormap('jet');
axis equal
hold on
vu_plot_n= vu_plot/norm(norm(vu)); % rescale
% vu_plot_n= vu_plot; % dont't rescale
femplot(x+vu_plot_n, conns, 'Color', 'k');
title("Stationary problem (check if normalized)" );
hold off
%--------------------------------------------------------------------------
toc

