%% DESCRIPTION
%--------------------------------------------------------------------------
% 03/03/2022 stationary study of fluid-solid rectangle connected top and
% bottom
% R. Liupekevicius Carnielli
%
% based com "compression_solution.m" from ACCM
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  - free          = traction = 0;
%  - fixed         = displac. = 0;
%  - open          = p=0;
%  - connect top and bottom elements
%  - f-s interface = continuity of normal velocity and stress;
%
%   (fixed).__________(connected)___________. (fixed)
%          |                                |
%          |           solid(tag=3)         |
%(periodic)|                                | (periodic)
%          |________ (f-s interface)________|
%          |                                |
%          |                                |
%  (open)  |           fluid(tag=2)         |  (open)
%          |                                |
%          |                                |
%          |________ (f-s interface)________|
%          |                                |
%(periodic)|           solid(tag=1)         | (periodic)
%          |                                |
%  (fixed).|__________(connected)___________|. (fixed)
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
% What are the low frequency modes ?
%
% Not reproducible in comsol.
%--------------------------------------------------------------------------
clear; close all;
tic

%% ADD TENSORLAB & SHARED FUNCTIONS & MESHES TO PATH
path(path,['C:\GitLab\Tools\TensorLab20\TensorLab20']);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\shared_functions"]);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\meshes"]);


toc
tic


%% PARAMETERS
disp('codeblock: PARAMETERS DEFINITION')
% geometric:
% H = 1; % height, in [m]
% L = 2; % length, in [m]
thick = 1;  % thickness (out-of-plane direction), in [m]


%--------------------------------------------------------------------------
% solid material properties
G        = 1000e6; % in [Pa]
kappa    = 3000e6; % in [Pa]
rhos     = 8000;   % in [kg/m3]
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% solid material properties PU (IN TERMS OF E AND nu)
% rhos     = 1000;   % in [kg/m3]
% E        = 1e6;    % [Pa]
% nu       = 0.4;    % []    
% G        = E/2/(1+  nu);
% kappa    = E/3/(1-2*nu);
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% fluid properties
rhof     = 1.225; % [kg/m3]
c        = 343 ;  % [m/s]
% pA       = 101325; [Pa]   % not used for order 1 equations.
%--------------------------------------------------------------------------


% define the basis
b  = {'e1'; 'e2'};
ee = cartesianbasis2d(b{1}, b{2});
e1 = ee(1);
e2 = ee(2);


% determine material coefficient tensor
I   = identity(2, ee);
I4  = identity(4, ee);
I4S = 1/2 * (I4 + rtranspose(I4));
C   = kappa *I*I + 2*G * (I4S - 1/3*I*I);

toc
tic

%% MANUALLY IMPORT COMSOL MESH AS STRUCT WITH CELL ELEMENTS
disp('codeblock: MESH')
%--------------------------------------------------------------------------
% open txt file comsol_mesh_solid_rect.mphtxt and copy/paste the node 
% position and connectivity matrix.
% -------------------------------------------------------------------------
%  Comments on 'conn' and 'tag' matrices for elements and edges      STATUS
% -------------------------------------------------------------------------
% - swap column 3 and 4 of conn ( quadrilateral element);                OK
% - no need for swaping when triangular element;                         OK
% - add 1 to all nodes of 'conn' (maybe 'tag' too). In comsol it starts
% in 0, whereas in matlab, it starts in 1;                               OK
% - attention to units! Use SI.                                          OK
% - check if the mesh has at least 6 nodes per wavelength          ON-GOING
%
%
%
% -------------------------------------------------------------------------
%  Comments on edge/element tags in comsol mesh file                                
% -------------------------------------------------------------------------
% - list of fluid-solid edge tag: 4,6
% This means that if an edge has tag 4 or 6, than the edge is at the fluid-
% solid interface.
% - list of element type (fluid or solid) after 'conn' matrix
%--------------------------------------------------------------------------

% load manually imported comsol mesh (use same mesh as in solid case)
load('es_fluidsolid_rect_MESH.mat');

% connect solid phase: top and bottom -> make connected sheet
    % solid
    % nodes on the bottom boundary of the mesh with corners
    bottom_s  = [1 3 6 10 15 21 28 36 45 54 63 72 81 90 99 108];
    % nodes on the top boundary of the mesh
    top_s     = [37 46 55 64 73 82 91 100 109 117 124 130 135 139 142 144]; 
    
    extra_e= 16-1;
    % loop on the edges to add volume elements
    for i=1:extra_e
        % element
        l.elem{2,1} = [l.elem{2,1}
                top_s(i) top_s(i+1) bottom_s(i+1) bottom_s(i)];
        % solid tag
        l.elem{2,2} = [l.elem{2,2}
                1]; 
    end

% copy to local variables (please double click on 'mesh' in workspace to
% see the meaning of each cell)
x    = l.x{2,2};    % tensor
conn = l.elem{2,1}; % conn matrix of elements

% plot mesh
figure(2)
femplot(x,conn,'Color' , 'black', 'Nodes', 'off', ...
'NodeNumbers' , 'on', 'ElementNumbers', 'on');

% mesh parameters
m  = size(conn, 1); % number of elements in the mesh
n  = size(x, 1);    % number of nodes in the mesh 


% SOLID
    % solid elements
    solid_elems = union(find(l.elem{2,2}==1),find(l.elem{2,2}==3)).';
    % position vector of the solid nodes
%     xs          = x(solid_elems);
    % conn solid
    conns       = conn(solid_elems,:);
    % solid nodes
    nodes_s     = unique(reshape(conns,1,[]));
    % number of solid elements in the mesh
    ms          = size(conns,1);
    % number of solid nodes
    ns          = length(nodes_s);


% FLUID
    % fluid elements
    fluid_elems = setdiff(1:m,solid_elems);
    % position vector of the fluid nodes
%     xf          = x(fluid_elems);
    % conn fluid
    connf       = conn(fluid_elems,:);
    % fluid nodes
    nodes_f     = unique(reshape(connf,1,[]));
    % number of fluid elements in the mesh
    mf          = size(connf,1);
    % number of fluid nodes
    nf          = length(nodes_f);

    
%% INTERFACE NODES (MANUALLY CHECKING TAGS)

% FORMER APPROACH
% import edge tags from comsol MANUALLY
tags_itr_edges=[4 6];
%(make 'for' loop below if list is too long), here only 2 interfaces

% 'interface edges' are the indices of 'conn' matrix
itr_edges_tag4 = find(l.edge{2,2}==4).';
itr_edges_tag6 = find(l.edge{2,2}==6).';
itr_edges  = [itr_edges_tag4 itr_edges_tag6];
% these are edges(two nodes per edge) that form the interface.
% The edge number ( or edge 'element') is the index of 'conn'

% mesh.edge('conn')( interface_edges , :)
conn_edges = l.edge{2,1}; %temporary assignment
conni_tag4 = conn_edges(itr_edges_tag4,:);
conni_tag6 = conn_edges(itr_edges_tag6,:);
conni      = conn_edges(itr_edges,:);
clear conn_edges;

% Get interface nodes (2D)
% - reshape conn (edges) matrix on a concatenated vector
% - keep 'unique' nodes of concatenated vector, i.e. remove duplicates
% - interface_nodes is therefore the index of elements' conn matrix
nodes_itr4     = unique(reshape( conni_tag4,1,[] ) ).';
nodes_itr6     = unique(reshape( conni_tag6,1,[] ) ).';
nodes_itr      = unique(reshape( conni     ,1,[] ) ).';


% ALTERNATIVE ( however need of 'conni' matrix)
% nodes_itr =  intersect(nodes_s,nodes_f);

%%  4 INTEGRATION POINTS WITHIN MASTER ELEMENT
%--------------------------------------------------------------------------
% Notation from TensorLab
% Linear quadrilateral element:
%         4-------3
%         | x   x |
%         |       |
%         | x   x |
%         1-------2
%      then CONN = [1 2 3 4].
% coordinates of integration points x
% (in the coordinate system of the master element)
% and the weight factors
%--------------------------------------------------------------------------

xi = [ -1/sqrt(3)*e1 - 1/sqrt(3)*e2
        1/sqrt(3)*e1 - 1/sqrt(3)*e2
        1/sqrt(3)*e1 + 1/sqrt(3)*e2
       -1/sqrt(3)*e1 + 1/sqrt(3)*e2 ];
w  = [ 1
       1
       1
       1 ];

toc
tic

%% ASSEMBLE SYSTEM MATRICES IN A SOLID ELEMENT LOOP 
disp('codeblock: ASSEMBLING')
%--------------------------------------------------------------------------
% Knn and Mnn are only temporary matrix and they are not the matrices K and
% M from discretized form in the notes. Note their size is the total number
% of nodes which include the fluid nodes. It's only relevant property is
% that it preserves the node indexing.
%--------------------------------------------------------------------------

Knn = zeros(2, ee, n, n); % zero matrix of nxn 2nd order tensors in basis ee

   for e = 1:size(conns,1) % loop over all solid elements

    iie = conns(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop
    Ke = zeros(2, ee, 4, 4); % zero matrix of 4x4 2nd order tensors
                             % in basis ee.
 
    for k = 1:length(w) % loop over 4 integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ 1/4*(1-xi1)*(1-xi2)
               1/4*(1+xi1)*(1-xi2)
               1/4*(1+xi1)*(1+xi2)
               1/4*(1-xi1)*(1+xi2) ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ -1/4*(1-xi2)*e1 - 1/4*(1-xi1)*e2
                      1/4*(1-xi2)*e1 - 1/4*(1+xi1)*e2
                      1/4*(1+xi2)*e1 + 1/4*(1+xi1)*e2
                     -1/4*(1+xi2)*e1 + 1/4*(1-xi1)*e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Ke = Ke + w(k) *  dot(gradNe, C, gradNe') * thick * det(J);
         
   
    end % end of interation point loop
   
   % assembly
    Knn(iie, iie) = Knn(iie, iie) + Ke;
   end % end of element loop

% assignment actual mass and stiffness matrices
K=Knn(nodes_s,nodes_s);


%% ASSEMBLE SYSTEM MATRICES IN A FLUID ELEMENT LOOP 
%--------------------------------------------------------------------------
% following the notation in my notes
% A - [1/Pa]      equivalent "mass matrix", term that multiplies d^2/dt^2 p 
% B - [1/(kg/m3)] equivalent "stiffness matrix", term that multiplies  p 
% d - [m/s2]      RHS term "external forces"
%--------------------------------------------------------------------------

Bnn = zeros(n, n); % zero matrix of nxn scalars


   for e = 1:size(connf,1)% loop over all fluid elements

    iie = connf(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop       
    Be = zeros(4, 4);        % zero matrix of 4x4 scalars

    for k = 1:length(w) % loop over integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ 1/4*(1-xi1)*(1-xi2)
               1/4*(1+xi1)*(1-xi2)
               1/4*(1+xi1)*(1+xi2)
               1/4*(1-xi1)*(1+xi2) ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ -1/4*(1-xi2)*e1 - 1/4*(1-xi1)*e2
                      1/4*(1-xi2)*e1 - 1/4*(1+xi1)*e2
                      1/4*(1+xi2)*e1 + 1/4*(1+xi1)*e2
                     -1/4*(1+xi2)*e1 + 1/4*(1-xi1)*e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Be = Be + w(k) *(1/rhof)*  dot(gradNe,gradNe') * thick * det(J);     
   
    end % end of interation point loop
   
% assembly
    Bnn(iie, iie) = Bnn(iie, iie) + Be;
   end % end of fluid element loop

% define actual stiffness and mass matrices.
B=Bnn(nodes_f,nodes_f);



%% INTEGRATION POINTS WITHIN MASTER EDGE (ELEMENT)
%--------------------------------------------------------------------------
% use here the same shape functions for both fluid and solid edges
% the edges have two nodes refered as 1 and 2 below.
%
%              solid              
%      1----x--------x-----2    
%              fluid             | (solid normal)   
%                                V 
% then CONN = [1 2].
% coordinates of integration points sketched as x.
%
% in the coordinate system of the master element and the weight factors
%--------------------------------------------------------------------------

xi = [ -1/sqrt(3)
        1/sqrt(3) ];
w  = [ 1
       1];

%% ASSEMBLE COUPLING MATRIX, LOOP ON INTERFACE EDGES 
%--------------------------------------------------------------------------
% use here the same shape functions for both fluid and solid elements
%--------------------------------------------------------------------------

Cnn = zeros(1, ee, n, n);        % zero matrix of nxn vectors


   for e = 1:size(conni,1) % loop over interface edges
    
    % edge e and its coordinates
    iie      = conni(e, :); % nodes of the current edge e
    xe       = x(iie);      % coordinates of these nodes
    
    % solid normal vector
    ve       = xe(2)-xe(1); % edge vector (of edge e)
    normal_s = ( dot(ve,e2)*e1 - dot(ve,e1)*e2)/...
                sqrt( dot(ve,e1)^2+dot(ve,e2)^2 ); % solid normal
%   normal_f = (-dot(ve,e2)*e1 + dot(ve,e1)*e2)/norm(ve); % fluid normal

    % compute element matrices in integration point loop
    Ce =  zeros(1, ee, 2, 2);        % 3x3 zero matrix of vectors

    for k = 1:length(w) % loop over 3 integration points 
        
       % xi(k) is the coordinate of the current integr. point
       % column of the shape functions
        Ne = [ 1/2*(1-xi(k))
               1/2*(1+xi(k)) ];

       % elementary length ds=|dx|d(xi)       =  1/2*norm(ve)*d(xi) 
       % elementary area   dA=|dx|d(xi) thick =  1/2*norm(ve)*thick*d(xi)
       
    
       % element matrix 
        Ce = Ce + w(k) * (Ne * Ne') * normal_s * 1/2*norm(ve)* thick ;
        
         
   
    end % end of interation point loop
   
% assembly
    Cnn(iie, iie) = Cnn(iie, iie) + Ce;

   end % end of interface edge loop

C = Cnn(nodes_s,nodes_f);

% % check nonzero elements seing the following matrices
% % "primitive conversion" to scalar for a matter of checking nonzero compon.
% Cnncheck = dot(Cnn,e1).^2 + dot(Cnn,e2).^2; 
% Cnncheck(nodes_itr,nodes_itr);
% Cnncheck(setdiff(1:n,nodes_itr),setdiff(1:n,nodes_itr));

toc
tic

%% NODE SORTING
disp('codeblock: NODE SORTING/ PBC PROJECTION/ PARTITIONING')
% Split the set of nodes 
% TODO: play with gmsh and automate the bc node sorting

% MANUALLY select nodes by looking at the femplot
% fluid
    % nodes on the left boundary of the mesh
    left_f    = [4 7 11 16 22]; 
    % nodes on the right boundary of the mesh
    right_f   = [123 129 134 138 141]; 
% solid
%     % nodes on the bottom boundary of the mesh
%     bottom_s  = [3 6 10 15 21 28 36 45 54 63 72 81 90 99];
%     % nodes on the top boundary of the mesh
%     top_s     = [46 55 64 73 82 91 100 109 117 124 130 135 139 142]; 
    % nodes on the left boundary of the mesh  (excluding corners)
    left_s    = [2 4 22 29]; 
    % nodes on the right boundary of the mesh  (excluding corners)
    right_s   = [116 123 141 143];
    % corner nodes
    corners_s = [1 108 144 37];

%% INDEX OF DISPLACEMENT/PRESSURE DOF IN NODES_S AND NODES_F
%--------------------------------------------------------------------------
% The matrices Mnn, Knn, Ann, Bnn, and Cnn have indices that correspond to
% the node number, i.g., line 1 are the terms related to node 1.
% The matrices M, K, A, B, and C, however, have indices that do not 
% correspond to the node number. Their indices correspond to the
% displacements degree of freedom or pressure degrees of freedom.
% Suppose, for example, you want to find the index in matrix M that
% correspond to the bottom-left corner displacement dof. This corner node
% is stored in the vector [corner], say first entry, corner(1). The index
% we are looking for, sometimes refered as 'new' index, is found using the
% following command:
% 
% find(nodes_s==corner(1)).
%
% get_ind() was implemented for this operation
%--------------------------------------------------------------------------

% fluid
    % left nodes
    doff_l = get_ind(nodes_f, left_f   ); 
    % right nodes
    doff_r = get_ind(nodes_f, right_f  ); 
    % interface nodes
    doff_i = get_ind(nodes_f, nodes_itr);
% solid 
%     % bottom nodes
%     dofs_b = get_ind(nodes_s, bottom_s ); 
%     % top nodes
%     dofs_t = get_ind(nodes_s, top_s    ); 
    % left nodes
    dofs_l = get_ind(nodes_s, left_s   ); 
    % right nodes
    dofs_r = get_ind(nodes_s, right_s  ); 
    % interface nodes
    dofs_i = get_ind(nodes_s, nodes_itr);
    % corner nodes
    dofs_c = get_ind(nodes_s, corners_s);

%--------------------------------------------------------------------------
% TODO:afterwards, check which sets (above) of dof were actually used.
%--------------------------------------------------------------------------

%% ELIMINATING DEPENDENT NODES OF SOLID PHASE
%--------------------------------------------------------------------------
% constrain relation of periodicity for 'old indices'
% u([top_s right_s])=u([bottom_s left_s])
%
% constrain relation of periodicity for 'new indices'
% u([dofs_t dofs_r])=u([dofs_b dofs_l])
%--------------------------------------------------------------------------

    % dependent dofs of periodic bc: u(dofs_de)= u(dofs_in)
    dofs_de = [dofs_r];   % [top right], excluding corners
    
    % independent nodes of periodic bc
    dofs_in = [dofs_l];    %[bottom left], excluding corners
    

    % safety check
    if(length(dofs_in)~= length(dofs_de)) 
        error("The periodic bc must connect same number of nodes");
    end

% unconstrained nodes ( nodes not at the boundary + corners)
% [dofs_c] in [dofs_un] because they will be prescribed later on.
dofs_un = setdiff(1:ns, [dofs_in dofs_de]); 

%--------------------------------------------------------------------------
% reordering mass, stiffness, and coupling matrix components as
% [dofs_un dofs_in dofs_de]
%--------------------------------------------------------------------------


% stiffness tensor matrix
% K=  Kuu Kui Kud   
%     Kiu Kii Kid  
%     Kdu Kdi Kdd  
Kuu=K(dofs_un, dofs_un); Kui=K(dofs_un, dofs_in); Kud=K(dofs_un, dofs_de);  
Kiu=K(dofs_in, dofs_un); Kii=K(dofs_in, dofs_in); Kid=K(dofs_in, dofs_de);  
Kdu=K(dofs_de, dofs_un); Kdi=K(dofs_de, dofs_in); Kdd=K(dofs_de, dofs_de);  
% coupling tensor matrix
% C=  Cu    
%     Ci   
%     Cd  
Cu=C(dofs_un,:);
Ci=C(dofs_in,:);
Cd=C(dofs_de,:);
%--------------------------------------------------------------------------    
% projection of reordered K and M on [uu;ui], i.e., eliminate ud.
% projection: uu = [T] uu   or matlab notation  [uu;ui;ud]= [T] [uu;ui];
%             ui       ui
%             ud    
% the reduced matrices are
% stiffness tensor matrix [K_] = [T]'[K][T]
% coupling tensor matrix  [C_] = [T]'[C]
% RHS tensor vector       [f_] = [T]'[f] (unknown on dirichlet nodes)
%                                        (kown on Neumman(=0)+free nodes)
%--------------------------------------------------------------------------

% projection matrix
T  = [                   eye(length([dofs_un dofs_in]))                  ;
       zeros(length(dofs_de),length(dofs_un))        eye(length(dofs_in))];

% reduced stiffness tensor matrix (dependent nodes eliminated)
K_ =  T'*[Kuu Kui Kud;
          Kiu Kii Kid;
          Kdu Kdi Kdd]*T;

% reduced mass tensor matrix (dependent nodes eliminated)
C_ =  T'*[Cu;
          Ci;
          Cd];


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% question: projection with T' or T.' ? TODO: test both
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%--------------------------------------------------------------------------
% set of remaining nodes, this is the vector that contains the rules for
% changing the indices from before projection to after projection.
dofs_re = [dofs_un dofs_in];
ns_re   = length(dofs_re);
%--------------------------------------------------------------------------

%% PRESCRIBED AND FREE NODES SPLIT
%--------------------------------------------------------------------------
% INDEX OF DISPLACEMENT DOF IN NODES_S (ELIMINATED DEPENDENT DOFS)
% same index change procedure must be executed but only in the solid phase.

%|__NOTES_NOTATION_|____ DESCRIPTION_______________________________________
%|                 |
%|        p        | prescribed nodes in the solid phase
%|_________________|_______________________________________________________
%|                 |
%|        f        | free nodes in the solid phase                                                      
%|                 |
%|_________________|_______________________________________________________
%|                 |
%|        p'       | prescribed nodes in the fluid phase
%|_________________|_______________________________________________________
%|                 |
%|        f'       | free nodes in the fluid phase
%|_________________|_______________________________________________________
%--------------------------------------------------------------------------

% solid
p  = get_ind(dofs_re,dofs_c);
f  = setdiff(1:ns_re,p);

% fluid
pp = [doff_l doff_r];
fp = setdiff(1:nf,pp);

%% PARTITIONING

% solid
    % stiffness tensor matrix
    % K =  Kpp Kpf   
    %      Kfp Kff 
    K_p_p   = K_(p , p ); K_p_f   = K_(p , f );
    K_f_p   = K_(f , p ); K_f_f   = K_(f , f );
    
% fluid 
    % 'stiffness' scalar matrix 
    % B =  Bp'p' Bp'f'   
    %      Bf'p' Bf'f' 
    B_pp_pp = B (pp, pp); B_pp_fp = B (pp, fp);
    B_fp_pp = B (fp, pp); B_fp_fp = B (fp, fp);

% coupling tensor matrix
    % C =  Cpp' Cpf'   
    %      Cfp' Cff' 
    C_p_pp  = C_(p , pp);C_p_fp  = C_(p , fp);
    C_f_pp  = C_(f , pp);C_f_fp  = C_(f , fp);

toc
tic


%% TRANFORM TENSOR MATRIX TO EQUIVALENT SCALAR MATRIX
disp('codeblock: TRANFORM TENSOR MATRIX TO EQUIVALENT SCALAR MATRIX')


% stiffness matrix Kff
sK_p_p = tm2sm_v( K_p_p, ee , 2 ); sK_p_f = tm2sm_v( K_p_f, ee , 2 );
sK_f_p = tm2sm_v( K_f_p, ee , 2 ); sK_f_f = tm2sm_v( K_f_f, ee , 2 );

% coupling matrix C
sC_p_pp = tm2sm_v(C_p_pp, ee, 1); sC_p_fp = tm2sm_v(C_p_fp, ee, 1);
sC_f_pp = tm2sm_v(C_f_pp, ee, 1); sC_f_fp = tm2sm_v(C_f_fp, ee, 1);

% zero matrices declaration directly as equivalent scalar matrix
%O_pp_p = sparse(zeros( length(pp) , size(ee,1)*length(p)));
O_fp_p = sparse(zeros( length(fp) , size(ee,1)*length(p)));
O_fp_f = sparse(zeros( length(fp) , size(ee,1)*length(f)));
O_pp_f = sparse(zeros( length(pp) , size(ee,1)*length(f)));

toc
tic

%% ASSEMBLE NONSYMMETRIC SYSTEM OF EQUATIONS 
disp('codeblock: SOLVE')
%--------------------------------------------------------------------------
% cond of the following matrices( with the term rhof*c2*) is lower
% % mu matrix
% muFF  = [  sK_f_f              sC_f_fp ;
%             O_fp_f          rhof*c^2*B_fp_fp    ];  
%--------------------------------------------------------------------------



% mu submatrices

% muPP  = [  sK_p_p              sC_p_pp ;
%             O_pp_p              B_pp_pp    ];  

% mu matrix
muPF  = [  sK_p_f              sC_p_fp ;
            O_pp_f              B_pp_fp    ];  
% mu matrix
muFP  = [  sK_f_p              sC_f_pp ;
            O_fp_p              B_fp_pp    ];  
% mu matrix
muFF  = [  sK_f_f              sC_f_fp ;
            O_fp_f              B_fp_fp    ];  


%% CONDITIONING OF ASSEMBLED SYSTEM
% fprintf('\n');
% fprintf('cond( muFF) = ');
% disp(condest(muFF));
% fprintf('\n');


%% PRESCRIBED DISPLACEMENT, PRESSURE AND ITS GRADIENTS

% reference position vector
xR = (   x(corners_s(1)) + x(corners_s(3))   )/2;

% ones
ones_u = ones(length(p) ,1);
ones_p = ones(length(pp),1);

% prescribed solid points are [corners_s]
    % MODEL INPUT
    uM     = 0*(e1+e2);
    graduM =  -0.3 *e1*e1  + 0.0  *e1*e2 +...
               0.0 *e2*e1  - 0.3  *e2*e2;
    % first-order comp. homogenization
    u_p    =  uM *ones_u + dot(graduM.', x(corners_s)-xR*ones_u);
    % transform to equivalent scalar (no need for sparsicity on a vector)
    su_p   = full(tm2sm_v(u_p, ee, 1));

if length(p)~=length(corners_s);         error('something wrong');end


% precribed fluid points are [left_f right_f]
    % MODEL INPUT
    pM     = -1;
    gradpM =  100*(e1+e2);
    % first-order comp. homogenization
    p_pp   =  pM *ones(length(pp),1) + ...
              dot(gradpM, x([left_f right_f])-xR*ones_p);
if length(pp)~=length([left_f right_f]); error('something wrong');end

% stack in a hybrid state vector
wP = [  su_p
         p_pp];

% %% SOLVE RIGHT/LEFT LINEAR SYSTEM OF EQUATIONS

% right stationary hybrid state vector
wr_f = - inv(muFF  ) * muFP   * wP;

% left stationary hybrid state vector
wl_f = - inv(muFF.') * muPF.' * wP;




% %% ASSIGN DISP. SOLUTION TO TENSOR VECTOR
%--------------------------------------------------------------------------
% 'ured' is  reduced displacement thanks
% to the elimitation of the dependend points due to periodic bc
%
%  length(ured) = length([uu;ui])
%  length(u   ) = length([uu;ui;ud])
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% free (reduced) displacements is the right stationary solution
    % declare
    ured_f = zeros(1, ee, length(f), 1);
    % assign
    % TODO: make the 'while' a function sv2tv() scalar vector to tensor vector
    % TODO: make the 'while' a function sm2tm() scalar matrix to tensor matrix
%     i=1;while i<=length(f)
%         ured_f(i,:) = wr_f(2*i-1)*e1+ wr_f(2*i)*e2;
%         fprintf("i = %d , 2*i-1 = %d ,2*i = %d \n", i,2*i-1,2*i)
% 
%         i=i+1;
%         end
    ii1=1:2:2*length(f)-1;
    ii2=2:2:2*length(f);
    ured_f = wr_f(ii1)*e1+ wr_f(ii2)*e2;

% total reduced displacements (indices ordered as [dofs_un dofs_in])
    % declare ured(ordered: [unconstrained independent] nodes)
    ured = zeros(1, ee, length(p)+length(f), 1);
    % assign
    % assign prescribed nodes
    ured(p,:) = u_p;
    % assign free nodes
    ured(f,:) = ured_f;
    % free unused variables
    clear ured_f; 
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% declare displacement tensor vectors (total # of dofs)
vu     = zeros(1, ee, ns, 1);
vu_aux = zeros(1, ee, ns, 1); 
% assign. Note vu_aux is in order [dofs_un dofs_in dofs_de]
vu_aux = T*ured;
% assign to vu. Note vu is in [nodes_s] order
vu([dofs_un dofs_in dofs_de])  = vu_aux;
% free unused variables
clear vu_aux;

% prepare for plot: length(vu_plot)>length(vu)
    % declare disp. vector for plot -  whole mesh
    vu_plot = zeros(1, ee, ns, 1);
    % assign vu to vector for plot, only solid nodes are non-zero
    vu_plot(nodes_s) = vu;
    % assign macro displacement to fluid nodes
%     vu_plot(nodes_f) = uM*ones(nf,1);

% %% ASSIGN PRESSURE SHAPE MODES TO SCALAR VECTOR vp

% extract pressure values from state vector wr_f
    % free pressures 
    p_fp = wr_f(end-length(fp)+1:end);

% declare pressure scalar vector
vp = zeros(nf,1);

% assign to scalar vector solution of size
vp(pp,:)=p_pp;
vp(fp,:)=p_fp;

% prepare for plot length(vp_plot)>length(vp)
    % declare pressure scalar vector for plot - whole mesh vector
    vp_plot = zeros(n,1);
    
    % assign vp to vector for plot
    vp_plot(nodes_f) = vp;

toc
tic


% %% PLOT PRESSURE MODE SHAPE 
disp('codeblock: PLOT')
% % select 'mode'and plot settings
% mode=3;
% figure(2)
% clf
% % plot the pressure solution
% scatter(dot(x,e1),dot(x,e2),400,vp_plot(:,mode),'filled')
% % colorbar('Ticks',[-1,-.5,0,.5 1]);
% colorbar;
% pmax = max(abs(vp_plot(:,mode)));
% caxis([-pmax pmax]);
% colormap('jet');
% axis equal
% hold on
% femplot(x, conn, 'Color', 'k');
% title("Mode shape of frequency "  + num2str(freqs(mode)) + " Hz" );
% hold off

% %% PLOT PRESSURE/DISP SHAPE 
figure(4)
clf

% plot the pressure solution and displacement
scatter(dot(x,e1),dot(x,e2),150,vp_plot,'filled')
colorbar;
pmax = max(abs(vp_plot));
caxis([-pmax pmax]);
colormap('jet');
axis equal
hold on
% vu_plot_n= vu_plot/norm(norm(vu)); % rescale
vu_plot_n= vu_plot; % dont't rescale
% femplot(x+vu_plot_n, conns, 'Color', 'k');
femplot(x+vu_plot_n, conns(1:end-extra_e,:), 'Color', 'k'); %remove back
title("Stationary problem" );
hold off


toc

