%% DESCRIPTION
%--------------------------------------------------------------------------
% 04/03/2022 stationary study of solid rectangle with quadratic elements
% R. Liupekevicius Carnielli
%
% based com "compression_solution.m" from ACCM
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  - periodic: disp.(source) = disp.(destination);
%  - fixed   : disp. = 0;
%
%        (u4).___(periodic)__.(u3)
%            |               |
% (periodic) |               | (periodic)
%            |               |
%            |___(periodic)__|
%        (u1)'               '(u2)
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Make sure you also have same element type in comsol.
%--------------------------------------------------------------------------
clear; close all;
tic
%% ADD TENSORLAB & SHARED FUNCTIONS & MESHES TO PATH
path(path,['C:\GitLab\Tools\TensorLab20\TensorLab20']);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\shared_functions"]);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\meshes"]);


%% PARAMETERS
% geometric:
% H = 1; % height, in [m]
% L = 2; % length, in [m]
thick = 1;  % thickness (out-of-plane direction), in [m]


%--------------------------------------------------------------------------
% solid material properties
G        = 1000e6; % in [Pa]
kappa    = 3000e6; % in [Pa]
rhos     = 8000;   % in [kg/m3]
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% solid material properties PU (IN TERMS OF E AND nu)
% rhos     = 1000;   % in [kg/m3]
% E        = 1e6;    % [Pa]
% nu       = 0.4;    % []    
% G        = E/2/(1+  nu);
% kappa    = E/3/(1-2*nu);
%--------------------------------------------------------------------------




% define the basis
b  = {'e1'; 'e2'};
ee = cartesianbasis2d(b{1}, b{2});
e1 = ee(1);
e2 = ee(2);


% determine material coefficient tensor
I   = identity(2, ee);
I4  = identity(4, ee);
I4S = 1/2 * (I4 + rtranspose(I4));
C   = kappa *I*I + 2*G * (I4S - 1/3*I*I);

%% MANUALLY IMPORT COMSOL MESH 
% open txt file comsol_mesh_solid_rect.mphtxt and copy/paste the node 
% position and connectivity matrix.
% -------------------------------------------------------------------------
%  Comments on 'conn' matrix and comsol mesh file                    STATUS
% -------------------------------------------------------------------------
% - swap column 3 and 4 of conn (quadrilateral) to set tensorlab format; OK
% - no need for swaping when triangular element;                         OK
% - add 1 to all nodes of 'conn'. In comsol it starts on 0, in matlab,1; OK
% - attention to units! Use SI.                                          OK
% - check if the mesh has at least 6 nodes per wavelength          ON-GOING
%--------------------------------------------------------------------------

% load manually imported comsol mesh
load('solid_rect_MESH.mat');

% transform to quadratic element
q=mesh2d_lin2qua(l);


% copy to local variables (please double click on 'mesh' in workspace to
% see the meaning of each cell)
x    = q.x{2,2};    % tensor
conn = q.elem{2,1}; % conn matrix of elements

% plot mesh
% figure(1)
% femplot(x,conn,'Color' , 'black', 'Nodes', 'off', ...
% 'NodeNumbers' , 'on', 'ElementNumbers', 'on');

% mesh parameters
m = size(conn, 1); % number of elements in the mesh
n = size(x, 1);    % number of nodes in the mesh 
    

%%  4 INTEGRATION POINTS WITHIN MASTER ELEMENT
%--------------------------------------------------------------------------
% Notation from TensorLab
% quadratic quadrilateral element:
%         4----7----3
%         |         |
%         8         6
%         |         |
%         1----5----2
%      then CONN = [1 2 3 4 5 6 7 8].
%
% The coordinates of integration points x (in the coordinate system of the
% master element). Exact integration up to polynomial of order 5.
%          --------- 
%         | x  x  x |
%         | x  x  x |
%         | x  x  x |
%          ---------
% and the weight factors
%--------------------------------------------------------------------------

xi = [ -sqrt(3/5) *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 + sqrt(3/5) *e2
       -sqrt(3/5) *e1 + sqrt(3/5) *e2
                0 *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 +         0 *e2       
                0 *e1 + sqrt(3/5) *e2                
       -sqrt(3/5) *e1 +         0 *e2
                0 *e1 +         0 *e2   ];
w  = [ 25/81
       25/81
       25/81
       25/81
       40/81
       40/81
       40/81
       40/81
       64/81     ];


%% ASSEMBLE SYSTEM MATRICES IN A SOLID ELEMENT LOOP 
%--------------------------------------------------------------------------
% Knn and Mnn are only temporary matrix and they are not the matrices K and
% M from discretized form in the notes. Note their size is the total number
% of nodes which include the fluid nodes. It's only relevant property is
% that it preserves the node indexing.
%--------------------------------------------------------------------------

K = zeros(2, ee, n, n); % zero matrix of nxn 2nd order tensors in basis ee

   for e = 1:size(conn,1) % loop over all solid elements

    iie = conn(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop
    Ke = zeros(2, ee, 8, 8); % zero matrix of 8x8 2nd order tensors
                             % in basis ee.
 
    for k = 1:length(w) % loop over 9 integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ -1/4*(1-xi1  )*(1-xi2  )*(1+xi1+xi2)
               -1/4*(1+xi1  )*(1-xi2  )*(1-xi1+xi2)
               -1/4*(1+xi1  )*(1+xi2  )*(1-xi1-xi2)
               -1/4*(1-xi1  )*(1+xi2  )*(1+xi1-xi2)
                1/2*(1-xi1^2)*(1-xi2  )
                1/2*(1+xi1  )*(1-xi2^2)
                1/2*(1-xi1^2)*(1+xi2  )
                1/2*(1-xi1  )*(1-xi2^2)           ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ - 1/4 * (-1 + xi2   )* (2*xi1 +   xi2) *e1 ...
                     - 1/4 * (-1 + xi1   )* (xi1   + 2*xi2) *e2
                     - 1/4 * (2*xi1 - xi2)* (-1    + xi2  ) *e1 ...
                     - 1/4 * (1 + xi1    )* (xi1   - 2*xi2) *e2
                       1/4 * (1 + xi2    )* (2*xi1 + xi2  ) *e1 ...
                     + 1/4 * (1 + xi1    )* (xi1   + 2*xi2) *e2
                       1/4 * (2*xi1 - xi2)* (1     + xi2  ) *e1 ...
                     + 1/4 * (-1 + xi1   )* (xi1 - 2*xi2  ) *e2
                                      xi1 * (-1 + xi2     ) *e1 ...
                                    + 1/2 * (-1 + xi1^2   ) *e2
                                      1/2 * ( 1 - xi2^2   ) *e1 ...
                                    - xi2 * ( 1 + xi1     ) *e2
                                    - xi1 * ( 1 + xi2     ) *e1 ...
                                    + 1/2 * ( 1 - xi1^2   ) *e2
                                      1/2 * (-1 + xi2^2   ) *e1 ...
                                    + xi2 * (-1 + xi1     ) *e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Ke = Ke + w(k) *  dot(gradNe, C, gradNe') * thick * det(J);
         
   
    end % end of interation point loop
   
   % assembly
    K(iie, iie) = K(iie, iie) + Ke;
   end % end of element loop




%% NODE SORTING
%spatial dimension
dim = size(ee,1);

% Manually select nodes by looking at the femplot
% nodes on the top boundary WITHOUT corners
[top,~] = find(q.x{2,1}(:,2)==1);
top = top(2:end-1)';
% top    = [25    28    30    35    39    43    48    53    58];

% nodes on the bottom boundary WITHOUT corners
bottom = [3     6    10    16    23    31    38    44    50];   
 % nodes on the left WITHOUT corners
left   = [2     4     8    12    18 ];         
% nodes on the right WITHOUT corners
right  = [56    57    59    60    61];      
% corner nodes on the boundary of the mesh
corners= [1 55 62 24];  % [1 2 3 4] which 1, 2 and 4 are prescribed 
% prescribed corners
pcorners = [corners(1) corners(2) corners(4)];

%% ELIMINATING DEPENDENT NODES

% constrain relation of periodicity 
% u(top)    = u(bottom) +  (u(coners(4))-u(coners(1))) ones()
% u(right)  = u(left)   +  (u(coners(2))-u(coners(1))) ones()

    % dependent nodes of periodic bc: u(iid)= u(iii)
    iid = [top right];   % excluding corners
    
    % independent nodes of periodic bc
    iii = [bottom left]; % excluding corners
    

    % safety check
    if(length(iii)~= length(iid)) 
        error("The periodic bc must connect same number of nodes");
    end

% unconstrained nodes ( nodes not at the boundary)
iiu = setdiff(1:n, [iii iid corners]); % exclude corners                           
iiu = [corners iiu];                   % include corners as first elements

% reordering mass and stiffness elements as [iiu iii iid]
    % stiffness tensor matrix
    % K_reordered =  Kuu Kui Kud   
    %                Kiu Kii Kid  
    %                Kdu Kdi Kdd  
    Kuu = K(iiu, iiu); Kui = K(iiu, iii); Kud = K(iiu, iid);  
    Kiu = K(iii, iiu); Kii = K(iii, iii); Kid = K(iii, iid);  
    Kdu = K(iid, iiu); Kdi = K(iid, iii); Kdd = K(iid, iid);  

% exclude corners again to stick with notes notation
iiu([1 2 3 4])=[]; 

% projection matrix components
O_1u = zeros(1,length(iiu));
O_u1 = O_1u';

O_1B = zeros(1,length(bottom));
O_B1 = O_1B';

O_1L = zeros(1,length(left));
O_L1 = O_1L';

O_uB = zeros(length(iiu),length(bottom));
O_Bu = O_uB';

O_uL = zeros(length(iiu),length(left));
O_Lu = O_uL';

O_BL = zeros(length(bottom),length(left));
O_LB = O_BL';

I_uu = eye(length(iiu),length(iiu));
I_BB = eye(length(bottom),length(bottom));
I_LL = eye(length(left),length(left));

I_B1 = ones(length(bottom),1);
I_L1 = ones(length(left),1);

% projection matrix
%     u_1   = T * u_1
%     u_2         u_2
%     u_3         u_4
%     u_4         u_un
%     u_un        u_in
%     u_in
%     u_de

T  = [  1       0      0      O_1u    O_1B    O_1L
        0       1      0      O_1u    O_1B    O_1L
       -1       1      1      O_1u    O_1B    O_1L
        0       0      1      O_1u    O_1B    O_1L
       O_u1    O_u1   O_u1    I_uu    O_uB    O_uL
       O_B1    O_B1   O_B1    O_Bu    I_BB    O_BL
       O_L1    O_L1   O_L1    O_Lu    O_LB    I_LL
      -I_B1    O_B1   I_B1    O_Bu    I_BB    O_BL
      -I_L1    I_L1   O_L1    O_Lu    O_LB    I_LL];


% reduced stiffness tensor matrix
K_ =  T'*[Kuu Kui Kud
          Kiu Kii Kid
          Kdu Kdi Kdd]*T;

% % reduced mass tensor matrix
% M_ =  T'*[Muu Mui Mud;
%           Miu Mii Mid;
%           Mdu Mdi Mdd]*T;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% question: projection with T' or T.' ? TODO: test both
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% set of remaining nodes
iire = [corners(1) corners(2) corners(4) iiu iii];

%% PRESCRIBED AND FREE NODES SPLIT
%--------------------------------------------------------------------------
% INDEX OF DISPLACEMENT DOF IN NODES_S (ELIMINATED DEPENDENT DOFS)
% same index change procedure must be executed but only in the solid phase.

%|__NOTES_NOTATION_|____ DESCRIPTION_______________________________________
%|                 |
%|        p        | prescribed nodes in the solid phase
%|_________________|_______________________________________________________
%|                 |
%|        f        | free nodes in the solid phase                                                      
%|                 |
%|_________________|_______________________________________________________
%--------------------------------------------------------------------------

% solid
    %first elements are the prescribed corners [1 2 4] by construction
    p  = get_ind(iire,pcorners); % = [1 2 3] 
    f  = setdiff(1:length(iire),p);



%% PARTITIONING

% solid
    % stiffness tensor matrix
    % K =  Kpp Kpf   
    %      Kfp Kff 
    K_p_p   = K_(p , p ); K_p_f   = K_(p , f );
    K_f_p   = K_(f , p ); K_f_f   = K_(f , f );
    
    

%% TRANFORM TENSOR MATRIX TO EQUIVALENT SCALAR MATRIX
disp('codeblock: TRANFORM TENSOR MATRIX TO EQUIVALENT SCALAR MATRIX')
tic

% stiffness matrix Kff
sK_p_p = tm2sm( K_p_p, ee , 2 ); sK_p_f = tm2sm( K_p_f, ee , 2 );
sK_f_p = tm2sm( K_f_p, ee , 2 ); sK_f_f = tm2sm( K_f_f, ee , 2 );


toc


%% PRESCRIBED DISPLACEMENT, PRESSURE AND ITS GRADIENTS

% reference position vector
% xR = (   x(corners(1)) + x(corners(3))   )/2;
xR = (   x(corners(1)) + x(corners(1))   )/2;



% scalar ones
ones_u = ones(length(p) ,1);


% prescribed solid points are [corners_s]
    % MODEL INPUT
    uM     = 0*(e1+e2);
    graduM = 0     *e1*e1   + 0 *e1*e2 +...
             0.3   *e2*e1   + 0 *e2*e2;
    % first-order comp. homogenization
    u_p    =  uM *ones_u + dot(graduM', x(pcorners)-xR*ones_u);
    % transform to equivalent scalar (no need for sparsicity on a vector)
    su_p   = full(tm2sm(u_p, ee, 1));




% %% SOLVE SYMMETRIC SYSTEM OF EQUATIONS

% right stationary hybrid state vector
u_f = - inv(sK_f_f  ) * sK_f_p   * su_p;





% %% ASSIGN DISP. SOLUTION TO TENSOR VECTOR
%--------------------------------------------------------------------------
% 'ured' is  reduced displacement thanks
% to the elimitation of the dependend points due to periodic bc
%
%  length(ured) = length([pcorners;uu;ui])
%  length(u   ) = length([pcorners;uu;ui;ud])
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% free (reduced) displacements is the right stationary solution
    % declare
    ured_f = zeros(1, ee, length(f), 1);
    % assign
    % TODO: make the 'while' a function sv2tv() scalar vector to tensor vector
    % TODO: make the 'while' a function sm2tm() scalar matrix to tensor matrix
    i=1;while i<=length(f)
        ured_f(i,:) = u_f(2*i-1)*e1+ u_f(2*i)*e2;
        i=i+1;
        end

% total reduced displacements (indices ordered as [dofs_un dofs_in])
    % declare ured(ordered: [unconstrained independent] nodes)
    ured = zeros(1, ee, length(p)+length(f), 1);
    % assign
    % assign prescribed nodes
    ured(p,:) = u_p;
    % assign free nodes
    ured(f,:) = ured_f;
    % free unused variables
    clear ured_f; 
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% declare displacement tensor vectors (total # of nodes)
vu     = zeros(1, ee, n, 1);
vu_aux = zeros(1, ee, n, 1); 
% assign. Note vu_aux is in order [iiu iii iid]
vu_aux = T*ured;
% assign to vu. Note vu is in [nodes] (original) order
vu([corners iiu iii iid])  = vu_aux;
% free unused variables
clear vu_aux;

% prepare for plot: 
    vu_plot = vu;
    


% %% PLOT DISP SHAPE (RESCALED DISP.)
figure(4)
clf
axis equal
hold on
% vu_plot_n= vu_plot/norm(norm(vu)); % rescale
vu_plot_n= vu_plot; % dont't rescale
femplot(x, conn, 'Color', 'r');
femplot(x+vu_plot_n, conn, 'Color', 'k');
title("Stationary problem" );
hold off


toc

