%% DESCRIPTION
%--------------------------------------------------------------------------
% 22/03/2022 stationary study of solid rectangle with quadratic elements
% R. Liupekevicius Carnielli
%
% based com "compression_solution.m" from ACCM
%
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%     (fixed)._____(free)____.(fixed)
%            |               |
%     (free) |               | (free)
%            |               |
%            |__(pressure)___|
%        
%%%%%%%%%%%%%%%%%%%%%% Boundary conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Make sure you also have same element type in comsol.
%--------------------------------------------------------------------------
clear; close all;

%% ADD TENSORLAB & SHARED FUNCTIONS & MESHES TO PATH
path(path,['C:\GitLab\Tools\TensorLab20\TensorLab20']);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\shared_functions"]);
path(path,["C:\GitLab\Tools\TensorLab20\TensorLab20"+...
            "\dev\meshes"]);


%% PARAMETERS
% geometric:
% H = 1; % height, in [m]
% L = 2; % length, in [m]
thick = 1;  % thickness (out-of-plane direction), in [m]


%--------------------------------------------------------------------------
% solid material properties
% G        = 1000e6; % in [Pa]
% kappa    = 3000e6; % in [Pa]
% rhos     = 8000;   % in [kg/m3]
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% solid material properties PU (IN TERMS OF E AND nu)
rhos     = 1000;   % in [kg/m3]
E        = 1e6;    % [Pa]
nu       = 0.4;    % []    
G        = E/2/(1+  nu);
kappa    = E/3/(1-2*nu);
%--------------------------------------------------------------------------




% define the basis
b  = {'e1'; 'e2'};
ee = cartesianbasis2d(b{1}, b{2});
e1 = ee(1);
e2 = ee(2);


% determine material coefficient tensor
I   = identity(2, ee);
I4  = identity(4, ee);
I4S = 1/2 * (I4 + rtranspose(I4));
C   = kappa *I*I + 2*G * (I4S - 1/3*I*I);

% pressure bc  p(x)= p0*x; 
p0=1; %[Pa]

%% MANUALLY IMPORT COMSOL MESH 
% open txt file comsol_mesh_solid_rect.mphtxt and copy/paste the node 
% position and connectivity matrix.
% -------------------------------------------------------------------------
%  Comments on 'conn' matrix and comsol mesh file                    STATUS
% -------------------------------------------------------------------------
% - swap column 3 and 4 of conn (quadrilateral) to set tensorlab format; OK
% - no need for swaping when triangular element;                         OK
% - add 1 to all nodes of 'conn'. In comsol it starts on 0, in matlab,1; OK
% - attention to units! Use SI.                                          OK
% - check if the mesh has at least 6 nodes per wavelength          ON-GOING
%--------------------------------------------------------------------------

% load manually imported comsol mesh
load('solid_rect_MESH.mat');

% transform to quadratic element
q=mesh2d_lin2qua(l);


% copy to local variables (please double click on 'mesh' in workspace to
% see the meaning of each cell)
x    = q.x{2,2};    % tensor
conn = q.elem{2,1}; % conn matrix of elements

% plot mesh
% figure(1)
% femplot(x,conn,'Color' , 'black', 'Nodes', 'off', ...
% 'NodeNumbers' , 'on', 'ElementNumbers', 'on');

% mesh parameters
m = size(conn, 1); % number of elements in the mesh
n = size(x, 1);    % number of nodes in the mesh 
    

%%  4 INTEGRATION POINTS WITHIN MASTER ELEMENT
%--------------------------------------------------------------------------
% Notation from TensorLab
% quadratic quadrilateral element:
%         4----7----3
%         |         |
%         8         6
%         |         |
%         1----5----2
%      then CONN = [1 2 3 4 5 6 7 8].
%
% The coordinates of integration points x (in the coordinate system of the
% master element). Exact integration up to polynomial of order 5.
%          --------- 
%         | x  x  x |
%         | x  x  x |
%         | x  x  x |
%          ---------
% and the weight factors
%--------------------------------------------------------------------------

xi = [ -sqrt(3/5) *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 + sqrt(3/5) *e2
       -sqrt(3/5) *e1 + sqrt(3/5) *e2
                0 *e1 - sqrt(3/5) *e2
        sqrt(3/5) *e1 +         0 *e2       
                0 *e1 + sqrt(3/5) *e2                
       -sqrt(3/5) *e1 +         0 *e2
                0 *e1 +         0 *e2   ];
w  = [ 25/81
       25/81
       25/81
       25/81
       40/81
       40/81
       40/81
       40/81
       64/81     ];


%% ASSEMBLE SYSTEM MATRICES IN A SOLID ELEMENT LOOP 
%--------------------------------------------------------------------------
% Knn and Mnn are only temporary matrix and they are not the matrices K and
% M from discretized form in the notes. Note their size is the total number
% of nodes which include the fluid nodes. It's only relevant property is
% that it preserves the node indexing.
%--------------------------------------------------------------------------

K = zeros(2, ee, n, n); % zero matrix of nxn 2nd order tensors in basis ee

   for e = 1:size(conn,1) % loop over all solid elements

    iie = conn(e, :); % nodes of the current element e
    xe  = x(iie);     % coordinates of these nodes
        
    % compute element matrices in integration point loop
    Ke = zeros(2, ee, 8, 8); % zero matrix of 8x8 2nd order tensors
                             % in basis ee.
 
    for k = 1:length(w) % loop over 9 integration points 
        
        xi1 = dot(xi(k), e1); % x1 coordinate of the current integr. point
        xi2 = dot(xi(k), e2); % x2 coordinate of the current integr. point

       % column of the shape functions
        Ne = [ -1/4*(1-xi1  )*(1-xi2  )*(1+xi1+xi2)
               -1/4*(1+xi1  )*(1-xi2  )*(1-xi1+xi2)
               -1/4*(1+xi1  )*(1+xi2  )*(1-xi1-xi2)
               -1/4*(1-xi1  )*(1+xi2  )*(1+xi1-xi2)
                1/2*(1-xi1^2)*(1-xi2  )
                1/2*(1+xi1  )*(1-xi2^2)
                1/2*(1-xi1^2)*(1+xi2  )
                1/2*(1-xi1  )*(1-xi2^2)           ];
        
        % column of the gradient of the shape functions
        % with respect to the local coordinates of the mater element
        gradxiNe = [ - 1/4 * (-1 + xi2   )* (2*xi1 +   xi2) *e1 ...
                     - 1/4 * (-1 + xi1   )* (xi1   + 2*xi2) *e2
                     - 1/4 * (2*xi1 - xi2)* (-1    + xi2  ) *e1 ...
                     - 1/4 * (1 + xi1    )* (xi1   - 2*xi2) *e2
                       1/4 * (1 + xi2    )* (2*xi1 + xi2  ) *e1 ...
                     + 1/4 * (1 + xi1    )* (xi1   + 2*xi2) *e2
                       1/4 * (2*xi1 - xi2)* (1     + xi2  ) *e1 ...
                     + 1/4 * (-1 + xi1   )* (xi1 - 2*xi2  ) *e2
                                      xi1 * (-1 + xi2     ) *e1 ...
                                    + 1/2 * (-1 + xi1^2   ) *e2
                                      1/2 * ( 1 - xi2^2   ) *e1 ...
                                    - xi2 * ( 1 + xi1     ) *e2
                                    - xi1 * ( 1 + xi2     ) *e1 ...
                                    + 1/2 * ( 1 - xi1^2   ) *e2
                                      1/2 * (-1 + xi2^2   ) *e1 ...
                                    + xi2 * (-1 + xi1     ) *e2 ];
        % Jacobian
        J = gradxiNe' * xe;
       
        % column of the gradient of the shape functions
        % with respect to the global coordinates of the mesh
        gradNe = dot(inv(J), gradxiNe);
        % element matrix and right hand side
        Ke = Ke + w(k) *  dot(gradNe, C, gradNe') * thick * det(J);
         
   
    end % end of interation point loop
   
   % assembly
    K(iie, iie) = K(iie, iie) + Ke;
   end % end of element loop



%% INTEGRATION POINTS WITHIN MASTER EDGE (ELEMENT)
%--------------------------------------------------------------------------
% use here the same shape functions for both fluid and solid edges
% the edges have three nodes refered as 1, 2 and 3 below.
%
%             exterior           
%      1--------3---------2  (master nodes)
%       --x-----x-----x--    (integration points)
%                                ^
%             solid              | (solid normal)   
%                                 
% then CONN = [1 3 2].
% coordinates of integration points sketched as x.  Exact integration up
% to polynomial of order 5.
%
% in the coordinate system of the master element and the weight factors
%--------------------------------------------------------------------------

xi = [ -sqrt(3/5)
              0
        sqrt(3/5) ];
w  = [ 5/9
       8/9
       5/9        ];
%% ASSEMBLE RHS TENSOR VECTOR, LOOP ON BOTTOM EDGES (TAG=1)

f_ext     = zeros(1, ee, n, 1);        % zero vector tensor of nx1
conni     = q.edge{2,1}([2 4 6 8 10],:);

   for e = 1:size(conni,1) % loop over interface edges
    
    % edge e and its coordinates
    iie      = conni(e, :); % nodes of the current edge e
    xe       = x(iie);      % coordinates of these nodes
    
    % solid normal vector
    ve       = xe(3)-xe(1); % edge vector (of edge e)
    normal_s = (-dot(ve,e2)*e1 + dot(ve,e1)*e2)/norm(ve); % normal

    % compute element matrices in integration point loop
    fe =  zeros(1, ee, 3, 1);        % 3x1 zero tensor of vector

    for k = 1:length(w) % loop over integration points 
        
       % xi(k) is the coordinate of the current integr. point(master elem)
       % column of the shape functions
        Ne = [ -1/2* (1-xi(k))*    xi(k)
                     (1-xi(k))* (1+xi(k))
                1/2*    xi(k) * (1+xi(k))  ];

      
       % integration point coordinate on physical edge element
       xk = Ne'*xe;
    
    
       % element matrix 
        fe = fe + w(k) * Ne* (-p0*dot(xk,e1) * normal_s) * 1/2*norm(ve)* thick ;
       % traction = (-p0*dot(x,e1))* normal_s
       % elementary length ds=|dx|d(xi)       =  1/2*norm(ve)*d(xi) 
       % elementary area   dA=|dx|d(xi) thick =  1/2*norm(ve)*thick*d(xi)
         
   
    end % end of interation point loop
   
% assembly
    f_ext(iie) = f_ext(iie) + fe;

   end % end of interface edge loop


%% NODE SORTING
%spatial dimension
dim = size(ee,1);

% Manually select nodes by looking at the femplot
% nodes on the top boundary WITHOUT corners
[top,~] = find(q.x{2,1}(:,2)==1);
top = top(2:end-1)';
% top    = [25    28    30    35    39    43    48    53    58];

% nodes on the bottom boundary WITHOUT corners
bottom = [3     6    10    16    23    31    38    44    50];   
 % nodes on the left WITHOUT corners
left   = [2     4     8    12    18 ];         
% nodes on the right WITHOUT corners
right  = [56    57    59    60    61];      
% corner nodes on the boundary of the mesh
corners= [1 55 62 24];  % [1 2 3 4] which 1, 2 and 4 are prescribed 
% prescribed corners
pcorners = [corners(3) corners(4)];


%% PRESCRIBED AND FREE NODES SPLIT
%--------------------------------------------------------------------------
% INDEX OF DISPLACEMENT DOF IN NODES_S (ELIMINATED DEPENDENT DOFS)
% same index change procedure must be executed but only in the solid phase.

%|__NOTES_NOTATION_|____ DESCRIPTION_______________________________________
%|                 |
%|        p        | prescribed nodes in the solid phase
%|_________________|_______________________________________________________
%|                 |
%|        f        | free nodes in the solid phase                                                      
%|                 |
%|_________________|_______________________________________________________
%--------------------------------------------------------------------------

% solid
    %first elements are the prescribed corners [1 2 4] by construction
    p  = pcorners; % = [1 2 3] 
    f  = setdiff(1:n,p);



%% PARTITIONING

% solid
    % stiffness tensor matrix
    % K =  Kpp Kpf   
    %      Kfp Kff 
    K_p_p   = K(p , p ); K_p_f   = K(p , f );
    K_f_p   = K(f , p ); K_f_f   = K(f , f );
    
    % external force (RHS)
    f_p     = f_ext(p);
    f_f     = f_ext(f);
    


%% PRESCRIBED DISPLACEMENT AND SOLVE

u_p = zeros(1,ee,length(pcorners),1);

% solve
u_f = K_f_f \ (f_f - dot(K_f_p, u_p));  

u    = zeros(1, ee, n, 1);        % zero vector tensor of nx1
u(p) = u_p;
u(f) = u_f;

% %% PLOT DISP SHAPE (RESCALED DISP.)
figure(4)
clf
axis equal
hold on
u_plot= u/norm(norm(u)); % rescale
femplot(x, conn, 'Color', 'r');
femplot(x+u_plot, conn, 'Color', 'k');
title("Stationary problem" );
hold off


figure(5)
clf
hold on
plot(dot(x([corners(1) bottom corners(2)]),e1), ...
     dot(u([corners(1) bottom corners(2)]),e1))
plot(dot(x([corners(1) bottom corners(2)]),e1), ...
     dot(u([corners(1) bottom corners(2)]),e2))
grid on;
box on;
legend('u','v');
hold off
